/*

 File: cimath.cpp, 2004/09/28

 CoStLy (COmplex interval STandard functions LibrarY), Version 1.0.0

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

//
//  Include header files
//
#include "cimath.h" //  Declaration of the complex functions
#include <limits>   //  From Standard Template Library
                    //  Contains information about the numeric type 'double'

using std::numeric_limits;

using Adapt::Interval;
using Adapt::PI;
using Adapt::HALFPI;
using Adapt::ZERO_INTERVAL;
using Adapt::ONE_INTERVAL;

#ifdef CXSC_VERSION

#include "dot.hpp" //"dotprecision" standard functions for CXSC version.

#endif

#include "error.h"

//-----------------------------------------------------------------------------
//
//  Preliminaries: Auxiliary functions and constants
//
//-----------------------------------------------------------------------------
//
//  Version():   CoStLy version number
//  max(x,y):    maximum of x and y (double x, y)
//  min(x,y):    minimum of x and y (double x, y)
//  SQRT_2:      interval constant containing sqrt(2)
//  INV_SQRT_2:  interval constant containing 1/sqrt(2)
//
//-----------------------------------------------------------------------------

//--  CoStLy version number ---------------------------------------------------
//
std::string CoStLy::Version()
{
  return "1.0.0";
}
//
//-- end CoStLy version number ------------------------------------------------

//-- max and min --------------------------------------------------------------
//
inline const double& max( const double& x, const double& y ) 
       { return ( x > y ? x : y ); }

inline const double& min( const double& x, const double& y ) 
       { return ( x < y ? x : y ); }
//
//-- end max and min ----------------------------------------------------------

//-- SQRT_2 and INV_SQRT_2 ----------------------------------------------------
//
inline const Interval& SQRT_2()
{
  static const Interval square_root_2( sqrt(Interval(2.0,2.0)) );
  return square_root_2;
}

inline const Interval& INV_SQRT_2()
{
  static const Interval inv_square_root_2( SQRT_2()/2.0 );
  return inv_square_root_2;
}
//
//-- end SQRT_2 and INV_SQRT_2 ------------------------------------------------

//-- err_msg -------------------------------------------------------- 040909 --
//
void err_msg( const std::string& f_string, const CInterval& z )
{
  std::cout<<"Farewell from CoStLy: " + f_string + " domain violation."<<std::endl;
  std::cout<<"Illegal argument z = "<<z<<std::endl<<std::endl;
  throw function_not_defined();
}
//
//-- end err_msg --------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//  Part I: Single-valued functions
//
//-----------------------------------------------------------------------------
//
//  Power operator  pow  is not listed here, because in this implementation,
//  pow  is based on the (multi-valued) logarithm.
//
//-----------------------------------------------------------------------------
//
//  Section 1: Separable functions
//             sqr, exp, cos, sin, cosh, sinh
//
//-----------------------------------------------------------------------------
//
//  The hyperbolic functions exp, sin, cos, sinh, cosh are separable:
//  Their real and imaginary parts are products of univariate real functions.
//
//
//  With Re(z)=x, Im(z)=y :
//
//       sqr   :   Re(sqr(z)) = sqr(x) - sqr(y)
//                 Im(sqr(z)) = 2 * x * y
//
//       exp   :   Re(exp(z)) = exp(x) * cos(y)
//                 Im(exp(z)) = exp(x) * sin(y)
//                                             
//       cos   :   Re(cos(z)) = cos(x) * cosh(y)
//                 Im(sin(x)) = -sin(x) * sinh(y)
// 
//       sin   :   Re(sin(z)) = sin(x) * cosh(y)
//                 Im(sin(x)) = cos(x) * sinh(y)
// 
//       cosh  :   Re(cosh(z)) = cosh(x) * cos(y)
//                 Im(cosh(z)) = sinh(x) * sin(y)
//                                                    
//       sinh  :   Re(sinh(z)) = sinh(x) * cos(y)
//                 Im(sinh(z)) = cosh(x) * sin(y)
//
//-----------------------------------------------------------------------------
 
//-- sqr ------------------------------------------------------------ 040602 --
//
CInterval CoStLy::sqr( const CInterval& z ) 
{
  Interval 
    rez = z.re(),
    imz = z.im();

  return CInterval( sqr( rez ) - sqr( imz ), 2.0 * rez * imz );
}
//
//-- end sqr ------------------------------------------------------------------

//-- exp ------------------------------------------------------------ 040625 --
//
CInterval CoStLy::exp( const CInterval& z ) 
{
  Interval 
    rez = exp( z.re() ),
    imz = z.im();

  return CInterval( rez * cos( imz ),  rez * sin( imz ) );
}
//
//-- end exp ------------------------------------------------------------------

//-- cos ------------------------------------------------------------ 040602 --
//
CInterval CoStLy::cos( const CInterval& z ) 
{
  Interval 
    rez = z.re(),
    imz = z.im();

  return CInterval( cos( rez ) * cosh( imz ) , -sin( rez ) * sinh( imz ) );
}
//
//-- end cos ------------------------------------------------------------------

//-- sin ------------------------------------------------------------ 040602 --
//
CInterval CoStLy::sin( const CInterval& z ) 
{
  Interval 
    rez = z.re(),
    imz = z.im();

  return CInterval( sin( rez ) * cosh( imz ) , cos( rez ) * sinh( imz ) );
} 
//
//-- end sin ------------------------------------------------------------------

//-- cosh ----------------------------------------------------------- 040602 --
//
CInterval CoStLy::cosh( const CInterval& z ) 
{
  Interval 
    rez = z.re(),
    imz = z.im();

  return CInterval( cos( imz ) * cosh( rez ) , sin( imz ) * sinh( rez ) );
}
//
//-- end cosh -----------------------------------------------------------------

//-- sinh ----------------------------------------------------------- 040602 --
//
CInterval CoStLy::sinh( const CInterval& z ) 
{
  Interval 
    rez = z.re(),
    imz = z.im();

  return CInterval( cos( imz ) * sinh( rez ) , sin( imz ) * cosh( rez ) );
}
//
//-- end sinh -----------------------------------------------------------------
 	
//-----------------------------------------------------------------------------
//
//  Section 2: tan, cot, tanh, coth
//
//  The implementation of cot, tanh, and coth is based on tan:
//
//  cot( z )  = tan( pi/2 - z )
//  tanh( z ) = transp( i * tan( transp( i * z ) ) 
//  coth( z ) = i * cot( i * z ) = i * tan( pi/2 - i * z )
//
//-----------------------------------------------------------------------------

//-- tan ------------------------------------------------------------ 040827 --
//
//  Complex tangent function
//
void horizontal_check( //-------------------------------------------- 040726 --
     const Interval& hy, const Interval& cosh_2y, double irez, double srez, 
     const Interval& hxl, const Interval& hxu, double& resxl, double& resxu )
//
//  Subroutine of tangent function.
//  Check intersections with extremal curves of tan on a horizontal boundary.
//  This subroutine is only called if an intersection occurs.
//  In this case, sinh( 2 * hy ) <> 0.0 (poles are handled before).
//
//  There may be 1 or 2 intersections.
//  If intersections lie next to a boundary of rez, then it is impossible to 
//  decide if there are 1 or 2 intersections.
//  In this case, 2 intersections are assumed 
//  (valid enclosure, at the expense of a potential slight overestimation).
//
{
  bool both = false, left = false, right = false;

  if( srez - irez > inf( PI() ) )
    //  2 intersections
    both = true;
  else 
    {
      Interval 
	res_l = cos( 2 * hxl ) - cosh_2y,
	res_u = cos( 2 * hxu ) - cosh_2y;

      if( inf( res_l * res_u ) > 0.0 )
	//  2 intersections
	both = true;
      else 
	{
	  if( sup( res_l * res_u ) < 0.0 )
	    {
	      //  1 intersection (3 intersections are PI() apart)
	      //  neither of the intervals res_l and res_u contains zero
	      if( inf( res_l ) > 0.0 )
		//  "left" intersection
		left = true;
	      else
		//  "right" intersection
		right = true;
	    }
	  else
	    //
	    //  1 (or both) intersections lie next to a boundary point
	    //  here, the task is to decide if 2 intersections occurs
	    //  if only 1 intersection takes place, then which one?
	    //
	    {
	      Interval 
		sin_2xl = sin( 2 * hxl ),
		sin_2xu = sin( 2 * hxu );

	      if( !disjoint( ZERO_INTERVAL(), res_l ) )
		//  intersection on the left boundary
		{
		  if( inf( sin_2xl ) >= 0.0 )
		    // "left" intersection
		    {
		      left = true;
		      //  remove the intersection by changing res_l!
		      res_l = - ONE_INTERVAL();
		    }
		  else
		    { 
		      if( sup( sin_2xl ) <= 0.0 )
			// "right" intersection
			{
			  right = true;
			  //  remove the intersection by changing res_l!
			  res_l =  ONE_INTERVAL();
			}
		      else
			//  zero is interior point of sin_2xl
			//  if the real sine function has optimal precision,
			//  this case should never happen
			both = true;
		    }
		}

	      if( !disjoint( ZERO_INTERVAL(), res_u ) )
		//  intersection on the right boundary
		{
		  if( inf( sin_2xu ) >= 0.0 )
		    // "left" intersection
		    {
		      left = true;
		      //  remove the intersection by changing res_u!
		      res_u = ONE_INTERVAL();
		    }
		  else
		    { 
		      if( sup( sin_2xu ) <= 0.0 )
			// "right" intersection
			{
			  right = true;
			  //  remove the intersection by changing res_u!
			  res_u = - ONE_INTERVAL();
			}
		      else
			//  zero is interior point of sin_2xu
			//  if the real sine function has optimal precision,
			//  this case should never happen
			both = true;
		    }
		}
	      //
	      //  finally, check if there is a second intersection
	      //
	      if( inf( res_l * res_u ) < 0.0 )
		both = true;
	    }
	}
    }    
  //
  //  Calculate extremal values
  //
  Interval re_tan = 1 / sinh( 2 * abs( hy ) );
  
  //  "left" intersection, sin( 2x ) > 0
  if( left or both )
    {
      resxl = min( resxl, inf( re_tan ) );
      resxu = max( resxu, sup( re_tan ) );
    }

  //  "right" intersection, sin( 2x ) < 0
  if( right or both )
    {
      resxl = min( resxl, - sup( re_tan ) );
      resxu = max( resxu, - inf( re_tan ) );
    }
}

CInterval CoStLy::tan( const CInterval& z ) //----------------------- 040827 --
{
  Interval
    rez = z.re(),
    imz = z.im();

  double 
    irez = inf(rez),
    srez = sup(rez),
    iimz = inf(imz),
    simz = sup(imz);

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double 
    resxl, resxu, resyl, resyu;
  //
  //  1st: check for poles
  //
  if( ( !disjoint( ZERO_INTERVAL(), imz ) ) and ( !disjoint( ZERO_INTERVAL(), cos( rez ) ) ) ) 
    err_msg( "tan", z );
  //
  //  2nd: real part
  //
  //  evaluate tan on vertical boundaries
  //
  Interval
    cos_2rez   = cos( 2 * rez ),
    sinh_imz_2 = sqr( sinh( imz ) );

  Interval
    re_tan_l = sin( 2 * hxl ) / ( 2 * ( sqr( cos( hxl ) ) + sinh_imz_2 ) ),
    re_tan_u = sin( 2 * hxu ) / ( 2 * ( sqr( cos( hxu ) ) + sinh_imz_2 ) );

  resxl = min( inf( re_tan_l ), inf( re_tan_u ) );
  resxu = max( sup( re_tan_l ), sup( re_tan_u ) );
  //
  //  look for extremal values on horizontal boundaries
  //  if one of the horizontal boundaries is the x-axes, 
  //  then the complex tangent is the real tangent on this 
  //  boundary, and due to monotonicity, its range 
  //  is already included in the ranges of the vertical boundaries
  //
  if( irez < srez )
    {
      Interval
	cosh_2yl = - 1 / cosh( 2 * hyl ),
	cosh_2yu = - 1 / cosh( 2 * hyu );

      if( !disjoint( cos_2rez, cosh_2yl ) and iimz != 0.0 )
	//extremal curve intersects lower boundary
	horizontal_check( hyl, cosh_2yl, irez, srez, hxl, hxu, resxl, resxu );
	  
      if( !disjoint( cos_2rez, cosh_2yu ) and simz != 0.0 )
	//extremal curve intersects upper boundary
	horizontal_check( hyu, cosh_2yu, irez, srez, hxl, hxu, resxl, resxu );
    }
  //
  //  3rd: imaginary part
  //
  //  evaluate tan on horizontal boundaries
  //
  Interval
    cos_rez_2 = sqr( cos( rez ) );

  Interval
    im_tan_l = sinh( 2 * hyl ) / ( 2 * ( cos_rez_2 + sqr( sinh( hyl ) ) ) ),
    im_tan_u = sinh( 2 * hyu ) / ( 2 * ( cos_rez_2 + sqr( sinh( hyu ) ) ) );

  resyl = min( inf( im_tan_l ), inf( im_tan_u ) );
  resyu = max( sup( im_tan_l ), sup( im_tan_u ) );

  //
  //  look for extremal values on vertical boundaries
  //  here, the situation is simpler than for the real part
  //  if 2 intersections with extremal curves appear , 
  //  one lies in the lower half plane, the other in the upper half plane
  //
  Interval
    cos_2xl = cos( 2 * hxl ),
    cos_2xu = cos( 2 * hxu );    
  Interval im_tan;

  if( iimz < 0.0 )
    //  intersection in lower half plane?
    {
      Interval 
	imz_h = Interval( iimz, min( simz, 0.0 ) ),
	cosh_2imz = - 1 / cosh( 2 * imz_h );
  
      if( ( !disjoint( cosh_2imz, cos_2xl ) ) )
	//extremal curve intersects left boundary
	//in this case, sin( 2 * xl ) <> 0.0 (no poles here!)
	{
	  im_tan = - 1 / abs( sin( 2 * hxl ) );
	  resyl = min( resyl, inf( im_tan ) );
	  resyu = max( resyu, sup( im_tan ) );
	}
      if( ( !disjoint( cosh_2imz, cos_2xu ) ) )
	//extremal curve intersects right boundary
	//in this case, sin( 2 * xu ) <> 0.0 (no poles here!)
	{
	  im_tan = - 1 / abs( sin( 2 * hxu ) );
	  resyl = min( resyl, inf( im_tan ) );
	  resyu = max( resyu, sup( im_tan ) );
	}
    }
  if( simz > 0.0 )
    //  intersection in upper half plane?
    {
      Interval 
	imz_h = Interval( max( iimz, 0.0 ), simz ),
	cosh_2imz = - 1 / cosh( 2 * imz_h );

      if( ( !disjoint( cosh_2imz, cos_2xl ) ) )
	//extremal curve intersects left boundary
	//in this case, sin( 2 * xl ) <> 0.0 (no poles here!)
	{
	  im_tan = + 1 / abs( sin( 2 * hxl ) );
	  resyl = min( resyl, inf( im_tan ) );
	  resyu = max( resyu, sup( im_tan ) );
	}
      if( ( !disjoint( cosh_2imz, cos_2xu ) ) )
	//extremal curve intersects right boundary
	//in this case, sin( 2 * xu ) <> 0.0 (no poles here!)
	{
	  im_tan = + 1 / abs( sin( 2 * hxu ) );
	  resyl = min( resyl, inf( im_tan ) );
	  resyu = max( resyu, sup( im_tan ) );
	}
    }
  
  return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );
  // return CInterval( Interval(  resxu-resxl, resxu-resxl ), Interval( resyu, resyu ) );
  //return CInterval( im_tan_l, im_tan_u );

}
//
//-- end tan ------------------------------------------------------------------

//-- cot ------------------------------------------------------------ 040730 --
//
//  cot( z )  = tan( pi/2 - z )
//
CInterval CoStLy::cot( const CInterval& z ) 
{
  return tan( HALFPI() - z );
}
//
//-- end cot ------------------------------------------------------------------

//-- tanh ----------------------------------------------------------- 040904 --
//
//  tanh( z ) = transp( i * tan( transp( i * z ) )
//
CInterval CoStLy::tanh( const CInterval& z ) 
{
  CInterval res = tan( CInterval( z.im(), z.re() ) ); 
  return CInterval( res.im(), res.re() );
}
//
//-- end tanh -----------------------------------------------------------------

//-- coth ----------------------------------------------------------- 040904 --
//
//  coth( z ) = i * cot( i * z ) = i * tan( pi/2 - i * z )
//
CInterval CoStLy::coth( const CInterval& z ) 
{
  CInterval zh = CInterval( z.im(), - z.re() );
  CInterval res = tan( HALFPI() + zh );  
  return( CInterval( - res.im(), res.re() ) );
}
//
//-- end coth -----------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Part II: Multi-valued functions
//
//-----------------------------------------------------------------------------
//
//  For these functions, two different procedures are implemented.
//
//  First, there is a subroutine for computing the principal value 
//  of the respective function. The principal value is usually
//  analytic in a subset of the complex plane and undefined otherwise. 
//
//  Second, there are procedures for computing interval supersets 
//  of all function values of the respective function, where feasible.
//
//-----------------------------------------------------------------------------
//
//  Section 1: Argument functions
//
//-----------------------------------------------------------------------------

//-- Arg: analytic argument function -------------------------------- 040916 --
//
//  (i)   Arg(Z) subset (-pi,pi).
//  (ii)  Arg([0,0]) = 0.
//  (iii) Arg(Z) is undefined if Z contains negative real numbers.
//  (iv)  Otherwise, Arg(Z) is the interval hull of { Arg(z) | z in Z, z<>0 }.
//
//  atan is the inverse function of tan(t), t in (-pi/2,pi/2).
//
Interval CoStLy::Arg( const CInterval& z ) 
{
  double 
    srez = sup( z.re() ),
    irez = inf( z.re() ), 
    simz = sup( z.im() ),
    iimz = inf( z.im() );

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double resl, resu;

  if( iimz > 0.0 )
    //  case I: Im(z) > 0
    {      
      resl = ( srez > 0.0 ? inf( atan( hyl/hxu ) ) : ( srez < 0.0 ? inf( atan( hyu/hxu ) + PI() ) : inf( HALFPI() ) ) ); 
      resu = ( irez > 0.0 ? sup( atan( hyu/hxl ) ) : ( irez < 0.0 ? sup( atan( hyl/hxl ) + PI() ) : sup( HALFPI() ) ) ); 
      return Interval( resl, resu );
    }
  else
    {
      if( simz < 0.0 )
	//  case II: Im(z) < 0
	{      
	  resl = ( irez < 0.0 ? inf( atan( hyu/hxl ) - PI() ) : ( irez > 0.0 ? inf( atan( hyl/hxl ) ) : - sup( HALFPI() ) ) ); 
	  resu = ( srez < 0.0 ? sup( atan( hyl/hxu ) - PI() ) : ( srez > 0.0 ? sup( atan( hyu/hxu ) ) : - inf( HALFPI() ) ) );  
	  return Interval( resl, resu );
	}
      else
	// 0 in Im(z)
	{
	  if( irez > 0.0 )
	    //  case III: Re(z) > 0 
	    //  z contains positive real values
	    {
	      resl = ( iimz < 0.0 ? inf( atan( hyl/hxl ) ) : 0.0 ); 
	      return Interval( resl, sup( atan( hyu/hxl ) ) );
	    }
	  else
	    //  z contains nonpositive real numbers 
	    {
	      if( irez < 0.0 )
		//  case IV: z contains negative real numbers
		err_msg( "Arg", z );
	      else
		//  case V: 0 in z, but z doesn't contain negative real numbers
		{
		  if( srez > 0.0 )
		    //  diam( Re(z) > 0.0 )
		    { 
		      resl = ( iimz < 0.0 ? - sup( HALFPI() ) : 0.0 );  
		      resu = ( simz > 0.0 ? sup( HALFPI() ) : 0.0 );
		      return Interval( resl, resu );
		    }
		  else
		    //  Re(z) == 0.0
		    {
		      if( iimz == 0.0 and simz == 0.0 )
			//  Z == 0
			return ZERO_INTERVAL();
		      else
			{
			  resl = ( iimz < 0.0 ? - sup( HALFPI() ) : inf( HALFPI() ) );
			  resu = ( simz > 0.0 ? sup( HALFPI() ) : - inf( HALFPI() ) );
			  return Interval( resl, resu );
			}
		    }
		}
	    }
	}
    }
}
//
//-- end Arg ------------------------------------------------------------------

//-- arg: non-analytic argument function ---------------------------- 040916 --
//
//  (i)   arg(Z) is defined for all Z in IC.
//  (ii)  arg(Z) subset [-pi,3*pi/2].
//  (iii) arg(Z) == Arg(Z) if Arg(Z) is well-defined.
//
//  atan is the inverse function of tan(t), t in (-pi/2,pi/2).
//
Interval CoStLy::arg( const CInterval& z ) 
{
  double 
    srez = sup( z.re() ),
    irez = inf( z.re() ), 
    simz = sup( z.im() ),
    iimz = inf( z.im() );

  double resl, resu;

  if( irez < 0.0 and iimz <= 0.0 and simz >= 0.0  )
    //  z contains negative real values
    {
      if( srez > 0.0 )
	//  0 in z and 0 interior point of Re(z)
	{
	  resl = ( iimz < 0.0 ? - sup( PI() ) : 0.0 );
	  resu = ( ( iimz < 0.0 and simz == 0.0 ) ? 0.0 : sup( PI() ) );
	  return Interval( resl, resu );
	}
      else
	{
	  if( iimz == simz )
	    //  z is real interval containing no positive values
	    return PI();
	  else
	    // sup( Re(z) ) <= 0, diam( Im(z) ) > 0
	    {
	      if( srez == 0.0 )
		{
		  resl = ( simz > 0.0 ? inf( HALFPI() ) : - sup( PI() ) );
		  resu = ( iimz < 0.0 ? ( simz > 0.0 ? sup( 3 * HALFPI() ) : - inf( HALFPI() ) ) : sup( PI() ) );
		  return Interval( resl, resu );
		}
	      else
		//   sup( Re(z) ) < 0, diam( Im(z) ) > 0
		{
		  Interval hyl(iimz), hyu(simz);
		  resl = ( simz > 0.0 ? inf( atan( hyu/srez ) + PI() ) : - sup( PI() ) );
		  resu = ( iimz < 0.0 ? ( simz > 0.0 ? sup( atan( hyl/srez ) + PI() ) : sup( atan( hyl/srez ) - PI() ) ) : sup( PI() ) );
		  return Interval( resl, resu );
		}
	    }
	}
    } 
  else
    //  Arg(z) is well-defined
    return  Arg( z );
}
//
//-- end arg ------------------------------------------------------------------

//-- arg_inclmon: non-analytic inclusion-monotone argument function - 040617 --
//  
//  (i)  arg_inclmon(Z) is defined for all Z in IC.
//  (ii) arg_inclmon(Z) = [-pi,pi] if Arg(Z) is not defined.
//
Interval CoStLy::arg_inclmon( const CInterval& z ) 
{
  if( inf( z.re() ) < 0.0 and inf( z.im() ) <= 0.0 and sup( z.im() ) >= 0.0  )
    return  Interval( -sup( PI() ), sup( PI() ) );
  else 
    return Arg( z );
}
//
//-- end arg_inclmon ----------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Section 2: Logarithms
//
//-----------------------------------------------------------------------------

//-- Ln: analytic natural logarithm --------------------------------- 040625 --
//
//  Ln(z) is undefined if z contains zero or negative real numbers.
//
CInterval CoStLy::Ln( const CInterval& z ) 
{
  if( inf(z.re()) <= 0.0 and inf(z.im()) <= 0.0 and sup(z.im()) >= 0.0  )
    //  z contains 0 or negative real values
    err_msg( "Ln", z );
  else
    return CInterval( ln( sqr( z.re() ) + sqr( z.im() ) ) / 2.0, Arg( z ) );
}
//
//-- end Ln -------------------------------------------------------------------

//-- ln: non-analytic natural logarithm ----------------------------- 040923 --
//
//  ln(z) is undefined if z contains zero.
//
CInterval CoStLy::ln( const CInterval& z ) 
{
  Interval abs_z_2 = sqr( z.re() ) + sqr( z.im() );
  if( inf( abs_z_2 ) == 0.0 )
    //  z contains 0
    err_msg( "ln", z );
  else
    return CInterval( ln( abs_z_2 ) / 2.0 , arg( z ) );
}
//
//-- end ln -------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Section 3: Root functions
//
//-----------------------------------------------------------------------------

//-- Sqrt: analytic square root ------------------------------------- 040626 --
//
Interval Re_Sqrt_point( const Interval& rez, const Interval& imz ) // 040626 --
//
//  Real part of analytic square root of A POINT INTERVAL ONLY.
//  Do not use this as a general function
//  - it's only a subroutine for Sqrt and sqrt_all.
//  The calculation is void if (rez+I*imz) is not a complex number.
//
{ 
  double 
    irez = inf( rez ),
    iimz = inf( imz );
 
  if( iimz == 0.0 )
    {
      if( irez >= 0.0 )
	return sqrt( rez );
      else
	return ZERO_INTERVAL();
    }
  else
    {
      Interval abs_z = sqrt( sqr( rez ) + sqr( imz ) );
      if( irez >= 0.0 )
	return INV_SQRT_2() * sqrt( abs_z + rez );
      else
	{ 
	  if( iimz > 0.0 )
	    return (  iimz * INV_SQRT_2() ) / sqrt( abs_z - rez );
	  else
	    return ( -iimz * INV_SQRT_2() ) / sqrt( abs_z - rez );
	}
    }
}

Interval Im_Sqrt_point( const Interval& rez, const Interval& imz ) // 040626 --
//
//  Imaginary part of analytic square root of A POINT INTERVAL ONLY
//  Do not use this as a general function
//  - it's only a subroutine for Sqrt and sqrt_all
//  The calculation is void if (rez+I*imz) is not a complex number 
//
{  
  double 
    irez = inf( rez ),
    iimz = inf( imz );
 
  if( iimz == 0.0 )
    {
      if( irez >= 0.0 )
	return ZERO_INTERVAL();
      else
	return sqrt( -rez );
    }
   else
    {
      Interval abs_z = sqrt( sqr( rez ) + sqr( imz ) );
      if( inf( rez ) >= 0.0 )
	return ( iimz * INV_SQRT_2() ) / sqrt( abs_z + rez );
      else
	{ 
	  if( iimz > 0.0 )
	    return INV_SQRT_2() * sqrt( abs_z - rez );
	  else
	    return - INV_SQRT_2() * sqrt( abs_z - rez );
	}
    }
}

CInterval CoStLy::Sqrt( const CInterval& z ) // --------------------- 040621 --
//
//  Analytic square root function
//  Sqrt(z) is undefined if z contains negative real numbers.
//
{
  double 
    irez = inf(z.re()),
    srez = sup(z.re()),
    iimz = inf(z.im()),
    simz = sup(z.im());
  Interval 
    hxl( irez ), hxu( srez ), hyl( iimz ), hyu( simz );
  double 
    resxl,  resxu, resyl, resyu;
  
  if( irez < 0.0 and iimz <= 0.0 and simz >= 0.0 )
    //  z contains negative real values
    err_msg( "Sqrt", z );
  else
    {  
      if( simz < 0.0 )
	{
	  //  z in lower half plane
	  //  min( Re ( Sqrt( z ) ) ) in upper left  corner
	  //  max( Re ( Sqrt( z ) ) ) in lower right corner
	  resxl = inf( Re_Sqrt_point( hxl, hyu ) );
	  resxu = sup( Re_Sqrt_point( hxu, hyl ) );
	  //  min( Im ( Sqrt( z ) ) ) in lower left  corner
	  //  max( Im ( Sqrt( z ) ) ) in upper right corner
	  resyl = inf( Im_Sqrt_point( hxl, hyl ) );
	  resyu = sup( Im_Sqrt_point( hxu, hyu ) );
	}
      else
	if( iimz > 0.0 )
	  {
	    //  z in upper half plane
	    //  min( Re ( Sqrt( z ) ) ) in lower left  corner
	    //  max( Re ( Sqrt( z ) ) ) in upper right corner
	    resxl = inf( Re_Sqrt_point( hxl, hyl ) );
	    resxu = sup( Re_Sqrt_point( hxu, hyu ) );
	    //  min( Im ( Sqrt( z ) ) ) in lower right corner
	    //  max( Im ( Sqrt( z ) ) ) in upper left  corner
	    resyl = inf( Im_Sqrt_point( hxu, hyl ) );
	    resyu = sup( Im_Sqrt_point( hxl, hyu ) );
	  }
	else
	  {
	    //  z intersects positive real axis
	    //  min( Re ( Sqrt( z ) ) ) on real line
	    //  max( Re ( Sqrt( z ) ) ) in lower or upper right corner
	    resxl = inf( sqrt( hxl ) );
	    resxu = ( - iimz > simz ? sup( Re_Sqrt_point( hxu, hyl ) ) : sup( Re_Sqrt_point( hxu, hyu ) ) ); 
	    //  min( Im ( Sqrt( z ) ) ) in lower left corner
	    //  max( Im ( Sqrt( z ) ) ) in upper left corner
	    resyl = inf( Im_Sqrt_point( hxl, hyl ) );
	    resyu = sup( Im_Sqrt_point( hxl, hyu ) );
	  }
      return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );
    }
}
//
//-- end Sqrt -----------------------------------------------------------------

//-- sqrt_all ------------------------------------------------------- 040621 --
//
//  sqrt_all(z) computes a list of 2 intervals containing all square roots of z
//
std::list<CInterval> CoStLy::sqrt_all( const CInterval& z ) 
{
  double 
    irez = inf(z.re()),
    srez = sup(z.re()),
    iimz = inf(z.im()),
    simz = sup(z.im());
  Interval 
    hxl( irez ), hxu( srez ), hyl( iimz ), hyu( simz );
  double 
    resxl,  resxu, resyl, resyu;
  CInterval w;
  
  if( irez < 0.0 and iimz <= 0.0 and simz >= 0.0 ) 
    //  z contains negative real values
    {
      if( iimz == 0.0 )
	//  z in upper half plane
	//  principal values can be used
	{
	  //  min( Re ( Sqrt( z ) ) ) in lower left  corner
	  //  max( Re ( Sqrt( z ) ) ) in upper right corner
	  resxl = inf( Re_Sqrt_point( hxl, hyl ) );
	  resxu = sup( Re_Sqrt_point( hxu, hyu ) );
	  //  min( Im ( Sqrt( z ) ) ) in lower right corner
	  //  max( Im ( Sqrt( z ) ) ) in upper left  corner
	  resyl = inf( Im_Sqrt_point( hxu, hyl ) );
	  resyu = sup( Im_Sqrt_point( hxl, hyu ) );
	}
      else
	{
	  if( simz == 0.0 )
	    //  z in lower half plane
	    //  principal values can be used in lower corners
	    {
	      //  z in lower half plane
	      //  min( Re ( Sqrt( z ) ) ) in upper left  corner
	      //  max( Re ( Sqrt( z ) ) ) in lower right corner
	      resxl = 0;                              ;
	      resxu = sup( Re_Sqrt_point( hxu, hyl ) );
	      //  min( Im ( Sqrt( z ) ) ) in lower left  corner
	      //  max( Im ( Sqrt( z ) ) ) in upper right corner
	      resyl = inf( Im_Sqrt_point( hxl, hyl ) );
	      resyu = ( srez > 0.0 ? 0.0 : -inf( sqrt( -hxu ) ) );
	    }
	  else
	    //  0 is interior point of Im( z )
	    {
	      if( srez <= 0.0 )
		{
		  //  0 is no interior point of Re (z )
		  //  in quadrant III,    Re( z ) = Im_Sqrt_point(|x|,y), 
		  //	                  Im( z ) = Re_Sqrt_point(|x|,y)
		  //  min( Re ( Sqrt( z ) ) ) in lower right corner = quadrant II/III
		  //  max( Re ( Sqrt( z ) ) ) in upper right corner = quadrant II
		  resxl = inf( Im_Sqrt_point( -hxu, hyl ) );
		  resxu = sup( Re_Sqrt_point( hxu, hyu ) );
		  //  min( Im ( Sqrt( z ) ) ) on real line
		  //  max( Im ( Sqrt( z ) ) ) in lower or upper left corner
		  resyl = inf( sqrt( -hxu ) );
		  resyu = ( - iimz > simz ? sup( Re_Sqrt_point( -hxl, hyl ) ) : sup( Im_Sqrt_point( hxl, hyu ) ) ); 
		}
	      else
		//  0 is interior point of z
		//  here, the principal values apply in all corner points
		{
		  resxl = 0;
		  //  max( Re ( Sqrt( z ) ) ) in lower or upper right corner
		  resxu = ( - iimz > simz ? sup( Re_Sqrt_point( hxu, hyl ) ) : sup( Re_Sqrt_point( hxu, hyu ) ) ); 
		  //  min( Im ( Sqrt( z ) ) ) in lower left corner
		  //  max( Im ( Sqrt( z ) ) ) in upper left corner
		  resyl = inf( Im_Sqrt_point( hxl, hyl ) );
		  resyu = sup( Im_Sqrt_point( hxl, hyu ) );
		}
	    }
	}
      w = CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );
    }
  else
    //  Sqrt( z ) is well-defined	    
    w = Sqrt( z );
      
  std::list<CInterval> res;
  res.push_back( w ); 
  res.push_back( -w );

  return res;
}
//
//-- end sqrt_all -------------------------------------------------------------

//-- Sqrt(z,n): analytic n-th root ---------------------------------- 040624 --
//
Interval Re_Sqrt_point( const Interval& rez, const Interval& imz, 
			unsigned int n ) // ------------------------- 040624 --
//
//  Real part of analytic n-th root.
//  Do not use this as a general function
//  - it's only a subroutine for Sqrt(z,n) and sqrt_all(z,n).
//  The calculation is validated but largely overestimating 
//  if (rez+I*imz) is not a complex number.
//
{   
  Interval abs_z_2 = sqr( rez ) + sqr( imz );
  if( sup( abs_z_2 ) == 0.0 )
    //  z == 0
    return ZERO_INTERVAL();
  else 
    return sqrt( abs_z_2, 2 * n ) * cos( CoStLy::Arg( CInterval( rez, imz ) ) / n );
}

Interval Im_Sqrt_point( const Interval& rez, const Interval& imz, 
			unsigned int n ) // ------------------------- 040624 --
//
//  Imaginary part of analytic n-th root.
//  Do not use this as a general function
//  - it's only a subroutine for Sqrt(z,n) and sqrt_all(z,n).
//  The calculation is validated but largely overestimating 
//  if (rez+I*imz) is not a complex number.
//
{ 
  Interval abs_z_2 = sqr( rez ) + sqr( imz );
  if( sup( abs_z_2 ) == 0.0 )
    //  z == 0
    return ZERO_INTERVAL();
  else 
    return sqrt( abs_z_2, 2 * n ) * sin( CoStLy::Arg( CInterval( rez, imz ) ) / n );
}

CInterval CoStLy::Sqrt( const CInterval& z, unsigned int n ) // ----- 040915 --
//
//  Analytic n-th root function
//  Sqrt(z,n) is undefined if z contains negative real numbers.
//
{
  if( n == 0 ) 
    return CInterval( ONE_INTERVAL() );
  if( n == 1 ) 
    return z;
  if( n == 2 ) 
    return Sqrt( z );
  else
    {
      double 
	irez = inf(z.re()),
	srez = sup(z.re()),
	iimz = inf(z.im()),
	simz = sup(z.im());
      Interval 
	hxl( irez ), hxu( srez ), hyl( iimz ), hyu( simz );
      double 
	resxl,  resxu, resyl, resyu;
      
      if( irez < 0.0 and iimz <= 0.0 and simz >= 0.0 ) 
	//  z contains negative real values
	err_msg( "Sqrt(z,n)", z );
      else
	{  
	  if( simz < 0.0 )
	    {
	      //  z in lower half plane
              CInterval hres = Sqrt( CInterval( z.re(), -z.im() ), n );
	      return CInterval( hres.re(), -hres.im() );
	    }
	  else
	    {
	      if( iimz > 0.0 )
		{
		  //  z in upper half plane
		  Interval tangle = tan( ( PI() * n ) / ( 2 * ( n-1 ) ) );
		  double tanglel = inf( tangle ), tangleu = sup( tangle );
		  //
		  //  min( Re( Root( z ) ) )
		  //
		  if ( irez >= 0.0  or  sup( hyl / irez ) <= tanglel )
		    //  lower boundary right of phi = n*Pi/(2n-2)
		    //  min( Re( Root( z ) ) ) in lower left corner
		    resxl = inf( Re_Sqrt_point( hxl, hyl, n ) );
		  else
		    {
		      if( srez < 0.0 and inf( hyl / srez ) >= tangleu )
			//  lower boundary left of phi = n*Pi/(2n-2)
			//  min( Re( Root( z ) ) ) in lower right corner
			resxl = inf( Re_Sqrt_point( hxu, hyl, n ) );
		      else
			//  lower boundary intersects phi = n*Pi/(2n-2)
			//  min( Re( Root( z ) ) ) on intersection
			resxl = inf( Re_Sqrt_point( iimz / tangle , hyl, n ) );
		    }
		  //
		  //  max( Re( Root( z ) ) )
		  //
		  if ( irez >= 0.0 or sup( hyu / irez ) <= tanglel )
		    //  upper boundary right of phi = n*Pi/(2n-2)
		    //  max( Re( Root( z ) ) ) in upper right corner
		    resxu = sup( Re_Sqrt_point( srez, simz, n ) );
		  else
		    {
		      if ( srez < 0.0 and inf( hyu / srez ) >= tangleu )
			//  upper boundary left of phi = n*Pi/(2n-2)
			//  max( Re( Root( z ) ) ) in upper left corner
			resxu = sup( Re_Sqrt_point( hxl, hyu, n ) );
		      else
			//  upper boundary intersects phi = n*Pi/(2n-2)
			//  max( Re( Root( z ) ) ) on upper left or right corner
			resxu = max( sup( Re_Sqrt_point( hxl, hyu, n ) ), sup( Re_Sqrt_point( hxu, hyu, n ) ) );
		    }
		  //
		  //  min( Im( Root( z ) ) )
		  //
		  if ( srez >= 0.0  or sup( hyl / srez ) <= tanglel )
		    //  right boundary right of phi = n*Pi/(2n-2)
		    //  min( Im( Root( z ) ) ) in lower right corner
		    resyl = inf( Im_Sqrt_point( hxu, hyl, n ) );
		  else
		    {
		      if( inf( hyu / srez ) >= tangleu )
			//  right boundary left of phi = n*Pi/(2n-2)
			//  min( Im( Root( z ) ) ) in upper right corner
			resyl = inf( Im_Sqrt_point( hxu, hyu, n ) );
		      else
			//  right boundary intersects phi = n*Pi/(2n-2)
			//  min( Im( Root( z ) ) ) on intersection
			resyl = inf( Im_Sqrt_point( hxu, srez * tangle, n ) );
		    }
		  //
		  //  max( Im( Root( z ) ) )
		  //      
		  if( irez >= 0.0 or sup( hyl / irez ) <= tanglel )
		    //  left boundary right of phi = n*Pi/(2n-2)
		    //  max( Im( Root( z ) ) ) in upper left corner
		    resyu = sup( Im_Sqrt_point( hxl, hyu, n ) );
		  else
		    {
		      if( inf( hyu / irez ) >= tangleu )
			//  left boundary left of phi = n*Pi/(2n-2)
			//  max( Im( Root( z ) ) ) in lower left corner
			resyu = sup( Im_Sqrt_point( hxl, hyl, n ) );
		      else
			//  left boundary intersects phi = n*Pi/(2n-2)
			//  max( Im( Root( z ) ) ) on lower or upper left corner
			resyu = max( sup( Im_Sqrt_point( hxl, hyl, n ) ), sup( Im_Sqrt_point( hxl, hyu, n ) ) );
		    }
		}
	      else
		{
		  //  z intersects positive real axis
		  //  min( Re( Root( z ) ) ) on real line
		  //  max( Re( Root( z ) ) ) in lower or upper right corner
		  resxl = ( irez == 0.0 ? 0.0 : inf( sqrt( hxl, n ) ) );
		  resxu = ( - iimz > simz ? sup( Re_Sqrt_point( hxu, hyl, n ) ) : sup( Re_Sqrt_point( hxu, hyu, n ) ) ); 
		  //  min( Im ( Sqrt( z ) ) ) in lower left corner
		  //  max( Im ( Sqrt( z ) ) ) in upper left corner
		  resyl = inf( Im_Sqrt_point( hxl, hyl, n ) );
		  resyu = sup( Im_Sqrt_point( hxl, hyu, n ) );
		}
	      return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );
	    }
	}
    }
}
//
//-- end Sqrt -----------------------------------------------------------------

//-- sqrt_all ------------------------------------------------------- 040628 --
//
std::list<CInterval> CoStLy::sqrt_all( const CInterval& z, unsigned int n )
//
//  sqrt_all(z,n) computes a list of n intervals containing all n-th roots of z
//
//  For n >=3, computing the optimal interval bounds is very expensive
//  and thus not considered cost-effective.
//
//  Hence, the polar form is used to calculate sqrt_all(z,n)
//  (observed overestimation of Re() and Im() in test cases: 5-15%).
//
//  z is enclosed into an interval in polar coordinates 
//  (i.e. a sector of an annulus), from which the roots 
//  (also polar intervals) are computed error-free 
//  (save roundoff, which is enclosed).
//
//  The the final inclusion of the roots into a rectangular complex interval
//  involves some additional overestimation.
//
{
  std::list<CInterval> res;

  if( n == 0 )
    {
      res.push_back( CInterval( ONE_INTERVAL(), ZERO_INTERVAL() ) );
      return res;
    }
  else if( n == 1 )
    {
      res.push_back(z);
      return res;
    }
  else if( n == 2 ) return sqrt_all( z );
  else
    {
      Interval 
	arg_z = arg( z ), root_abs_z = sqrt( abs( z ), n );

      for(int k = 0; k < n; k++)
	{
	  Interval arg_k = ( arg_z + 2 * k * PI() ) / n;

	  res.push_back( CInterval( root_abs_z * cos( arg_k ),
				    root_abs_z * sin( arg_k ) ) );
	}
      return res;
    }
}
//
//-- end sqrt_all -------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Section 4: Power functions
//
//-----------------------------------------------------------------------------

//-- power_fast ----------------------------------------------------- 040714 --
//
//  Fast, validated power function for integer powers, based on exp and ln.
//  Medium amount of overestimation.
//
CInterval CoStLy::power_fast( const CInterval& z, int n ) 
{
  if( n == 0 )
    return CInterval( ONE_INTERVAL() );
  else if( n == 1 ) 
    return z;
  else if( n == -1 ) 
    return 1 / z;
  else if( n == 2 ) 
    return sqr( z );
  else 
    //  n >= 3  or  n <= -2
    //  If z is a large interval, z^n is a distorted set, for which the 
    //  inclusion into a complex rectangle contains large overestimation. 
    //  For example, if  n * Arg( z )  intersects the cartesian axes 
    //  more than once then  0  is contained in the rectangular enclosure 
    //  of z^n. 
    //  For n <= -2, also inversion of z or z^n is required; 
    //  both inversions lead to large overestimation of the resulting interval.
    //
    //  The computation of an optimal rectangular enclosure is implemented
    //  in power(z,n). In power_fast(z,n), z is enclosed into a sector S of 
    //  an annulus. S^n is again some sector S' of a (different) annulus. 
    //  S' is calculated exactly (apart from roundoff), and then enclosed 
    //  into a rectangle. There is a certain amount of overestimation  
    //  compared with the optimal rectangular enclosure of z^n, but the 
    //  calculations are much cheaper here.
    {
      Interval abs_z = abs( z );

      if( n < 0 and inf( abs_z ) == 0.0 )
	//  z contains 0 
	err_msg( "power_fast", z );

      if( sup( abs_z ) == 0.0 )
        return CInterval( 0.0, 0.0 );
      else
        { 
          Interval arg_z = arg( z ); 
	  Interval abs_z_n = exp( n * ln( abs_z ) );
	  
	  return CInterval( abs_z_n * cos( n * arg_z ),
                            abs_z_n * sin( n * arg_z ) ); 
	}
    }
}
//
//-- end power_fast -----------------------------------------------------------

//-- power ---------------------------------------------------------- 040720 --
//
//  Power function for integer powers with optimal (save roundoff) accuracy.
//
//  The extremal curves of Re(z^n) and Im(z^n) are straight lines defined
//  by the rays  arg(z) = k pi/ ( 2n - 2 ), k = 0,...4n-5.
//
//  Intersections of these rays with the boundary of z are called 
//  "ray intersections" in the following.
//
CInterval power_point( const CInterval& z, int n ) //---------------- 040715 --
//
//  z^n for A POINT INTERVAL z ONLY.
//  Do not use this as a general function
//  - it's only a subroutine for power.
//  The calculation may break down otherwise.
//  The case 0 in z for negative n is handled in power(z,n).
//
{        
  if( inf( z.re() ) == 0.0 and inf( z.im() ) == 0.0 )
    //  if z is a valid point interval, it must be 0
    return CInterval( ZERO_INTERVAL(), ZERO_INTERVAL() );
  else
    {
      Interval abs_z = abs( z );
      Interval arg_z = CoStLy::arg( z ); 
      Interval abs_z_n = exp( n * ln( abs_z ) );
      
      return CInterval( abs_z_n * cos( n * arg_z ),
			abs_z_n * sin( n * arg_z ) );  
    }
}

void update_res( const CInterval& res, //---------------------------- 040716 --
                 double& resxl, double& resxu, double& resyl, double& resyu )
//  Subroutine of power(z,n).
//  Update extremal values of power function.
{
  resxl = min( resxl, inf( res.re() ) );
  resxu = max( resxu, sup( res.re() ) );
  resyl = min( resyl, inf( res.im() ) );
  resyu = max( resyu, sup( res.im() ) );
}

void horizontal_check( //-------------------------------------------- 040720 --
     const Interval& hy, double hyy, const Interval& arg_h, 
     double irez, double srez, 
     double& resxl, double& resxu, double& resyl, double& resyu, int n )
//  Subroutine of power(z,n).
//  Check all relevant ray intersections on a horizontal boundary.
{
  double r_int;
  int
    il1, il2, iu1, iu2;
  CInterval res;
  //  
  //  Here, the most difficult task is 
  //  to determine the relevant ray intersections 
  //  Both arg_hl and n can be negative, which makes it very complicated 
  //  to decide the correct indexes for the "rightmost" intersections, etc.
  //
  //  To simplify the distinction of cases, we introduce the variable 
  //  nofrays (number of rays) = abs( n-1 )
  //
  //  Note that we still have to pass n to power_point
  //
  int nofrays = abs( n-1 );
  double arg_hl = inf( 2 * nofrays * arg_h / PI() );
  if( arg_hl >= 0.0 )
    modf( arg_hl, &r_int );
  else
    modf( arg_hl-1, &r_int );
  il1 = int( r_int + 1.0 );
  double arg_hu = sup( 2 * nofrays * arg_h / PI() );
  if( arg_hu >= 0.0 )
    modf( arg_hu, &r_int );
  else
    modf( arg_hu-1, &r_int );
  iu1 = int( r_int );

  if( iu1 >= il1 )
    //  at least one ray intersection
    //  maybe more?
    {
      if( iu1 > il1 + 3 ) 
	//
	//  we're in trouble: 
	//  there are more than 4 ray intersections
	//  now we must decide which of these are relevant
	//  depending on il1, iu1, n and on the quadrants, 
	//  4 to 7 intersections must be considered
	//
	if( n > 0 )
	  //  outmost intersections are relevant
	  {
	    if( irez >= 0.0 ) 
	      //  z in quadrants I or IV: 
	      //  only 4 rightmost intersections are relevant
	      {  
		if( hyy < 0.0 )
		  //  quadrant IV
		  il1 = iu1 - 3;
		else
		  //  quadrant I
		  iu1 = il1 + 3;
	      }
	    else if( srez <= 0.0 )
	      //  z in quadrants II or III:
	      //  only 4 leftmost intersections are relevant
	      {  
		if( hyy > 0.0 )
		  //  quadrant II
		  il1 = iu1 - 3;
		else
		  //  quadrant III
		  iu1 = il1 + 3;
	      }
	    else
	      //  z intersects imaginary axes
	      //  we may have two lists of intersections!
	      {  
		iu2 = iu1;
		il2 = iu2 - 3;
		iu1 = il1 + 3;
		//  remove intersection points that are contained in both lists
		if( il2 <= iu1 )
		  il2 = iu1 + 1;
		//
		//  here, list 2 is processed
		//  list 1 is processed below
		//
		for(int i = il2; i <= iu2; i++)
		  {
		    Interval cotangle = cot( ( PI() * i ) / ( 2 * nofrays ) );
		    res = power_point( CInterval( hy * cotangle , hy ), n );
		    update_res( res, resxl, resxu, resyl, resyu );
		  }
	      }
	  }
	else
	  //  n < 0:
	  //  innermost intersections are relevant
	  {
	    if( irez >= 0.0 ) 
	      //  z in quadrants I or IV: 
	      //  only 4 leftmost intersections are relevant
	      {  
		if( hyy > 0.0 )
		  //  quadrant I
		  il1 = iu1 - 3;
		else
		  //  quadrant IV
		  iu1 = il1 + 3;
	      }
	    else if( srez <= 0.0 )
	      //  z in quadrants II or III:
	      //  only 4 rightmost intersections are relevant
	      {  
		if( hyy < 0.0 )
		  //  quadrant III
		  il1 = iu1 - 3;
		else
		  //  quadrant II
		  iu1 = il1 + 3;
	      }
	    else
	      //  z intersects imaginary axes
	      //  we have one lists of 5 to 7 intersections
	      {
		if( hyy > 0.0 )
		  {
		    il2 = nofrays - 3;
		    iu2 = nofrays + 3;
		  }
		else
		  {
		    il2 = -nofrays - 3;
		    iu2 = -nofrays + 3;
		  }
		//  list 1 contains all intersections
		//  list 2 is a filter for all relevant intersections
		//  store intersection of lists 1 and 2 in list 1
		il1 = ( il1 > il2 ? il1 : il2 );
		iu1 = ( iu1 < iu2 ? iu1 : iu2 );
	      }
	  }
      //
      //  list 1 has been left for processing
      //
      for(int i = il1; i <= iu1; i++)
	{
	  Interval cotangle = cot( ( PI() * i ) / ( 2 * nofrays ) );
	  res = power_point( CInterval( hy * cotangle , hy ), n );
	  update_res( res, resxl, resxu, resyl, resyu );
	}
    }
}

void vertical_check( //---------------------------------------------- 040720 --
     const Interval& hx, double hxx, const Interval& arg_h,
     double iimz, double simz, 
     double& resxl, double& resxu, double& resyl, double& resyu, int n )
//  Subroutine of power(z,n).
//  Check all relevant ray intersections on a vertical boundary.
{
  double r_int;
  int
    il1, il2, iu1, iu2;
  CInterval res;
  //  
  //  Here, the most difficult task is 
  //  to determine the relevant ray intersections 
  //  Both arg_hl and n can be negative, which makes it very complicated 
  //  to decide the correct indexes for the topmost intersections, etc.
  //
  //  To simplify the distinction of cases, we introduce the variable 
  //  nofrays (number of rays) = abs( n-1 )
  //
  //  Note that we still have to pass n to power_point
  //
  int nofrays = abs( n-1 );

  double arg_hl = inf( 2 * nofrays * arg_h / PI() );
  if( arg_hl >= 0.0 )
    modf( arg_hl, &r_int );
  else
    modf( arg_hl-1, &r_int );
  il1 = int( r_int + 1.0 );
  double arg_hu = sup( 2 * nofrays * arg_h / PI() );
  if( arg_hu >= 0.0 )
    modf( arg_hu, &r_int );
  else
    modf( arg_hu-1, &r_int );
  iu1 = int( r_int );
 
  if( iu1 >= il1 )
    //  at least one ray intersection
    //  maybe more?
    {
      if( iu1 > il1 + 3 ) 
	//
	//  we're in trouble: 
	//  there are more than 4 ray intersections
	//  now we must decide which of these are relevant
	//  depending on il1, iu1, n and on the quadrants, 
	//  4 to 7 intersections must be considered
	//
	if( n > 0 )
	  //  outmost intersections are relevant
	  {
	    if( iimz >= 0.0 ) 
	      //  z in quadrants I or II: 
	      //  only 4 topmost intersections are relevant
	      {  
		if( hxx > 0.0 )  
		  //  quadrant I
		  il1 = iu1 - 3;
		else
		  //  quadrant II
		  iu1 = il1 + 3;
	      }
	    else if( simz <= 0.0 )
	      //  z in quadrants III or IV:
	      //  only 4 lowest intersections are relevant
	      {  
		if( hxx < 0.0 ) 
		  //  quadrant III
		  il1 = iu1 - 3;
		else
		  //  quadrant IV
		  iu1 = il1 + 3;
	      }
	    else
	      //  z intersects real axes
	      //  we may have two lists of intersections!
	      {  
		iu2 = iu1;
		il2 = iu2 - 3;
		iu1 = il1 + 3;
		//  remove intersection points that are contained in both lists
		if( il2 <= iu1 )
		  il2 = iu1 + 1;
		//
		//  here, list 2 is processed
		//  list 1 is processed below
		//
		for(int i = il2; i <= iu2; i++)
		  {
		    Interval tangle = tan( ( PI() * i ) / ( 2 * nofrays ) );
		    res = power_point( CInterval( hx, hx * tangle ), n );
		    update_res( res, resxl, resxu, resyl, resyu );
		  }
	      }
	  }
	else
	  //  n < 0:
	  //  innermost intersections are relevant
	  {
	    if( iimz >= 0.0 ) 
	      //  z in quadrants I or II: 
	      //  only 4 lowest intersections are relevant
	      {  
		if( hxx < 0.0 )
		  //  quadrant II
		  il1 = iu1 - 3;
		else
		  //  quadrant I
		  iu1 = il1 + 3;
	      }
	    else if( simz <= 0.0 )
	      //  z in quadrants III or IV:
	      //  only 4 topmost intersections are relevant
	      {  
		if( hxx > 0.0 )
		  //  quadrant IV
		  il1 = iu1 - 3;
		else
		  //  quadrant III
		  iu1 = il1 + 3;
	      }
	    else
	      //  z intersects real axes
	      //  we have one lists of 5 to 7 intersections
	      {
		if( hxx > 0.0 )
		  {
		    il2 = -3;
		    iu2 =  3;
		  }
		else
		  {
		    il2 = 2 * nofrays - 3;
		    iu2 = 2 * nofrays + 3;
		  }
		//  list 1 contains all intersections
		//  list 2 is a filter for all relevant intersections
		//  store intersection of lists 1 and 2 in list 1
		il1 = ( il1 > il2 ? il1 : il2 );
		iu1 = ( iu1 < iu2 ? iu1 : iu2 );
	      }
	  }
      //
      //  list 1 has been left for processing
      //
      for(int i = il1; i <= iu1; i++)
	{
	  Interval tangle = tan( ( PI() * i ) / ( 2 * nofrays ) );
	  res = power_point( CInterval( hx, hx * tangle ), n );
	  update_res( res, resxl, resxu, resyl, resyu );
	}
    }
}

CInterval CoStLy::power( const CInterval& z, int n ) //-------------- 040720 --
//
//  Power function for integer powers with optimal (save roundoff) accuracy.
//
{
  if( n == 0 )
    return CInterval( ONE_INTERVAL() );
  else if( n == 1 ) 
    return z;
  else if( n == -1 ) 
    return 1 / z;
  else if( n == 2 ) 
    return sqr( z );
  else 
    //  n >= 3  or  n <= -2
    //
    //  n may have a large absolute value and z may be a large interval. 
    //  In this case, we calculate function values at specific points
    //  and look for min and max.
    //  
    //  An implementation with fewer function evaluations would be possible,
    //  at the expense of an unreadable source code.
    {
      Interval abs_z = abs( z );

      if( n < 0 and inf( abs_z ) == 0.0 )
	//  z contains 0 
	err_msg( "power", z );

      if( sup( abs_z ) == 0.0 )
        return CInterval( 0.0, 0.0 );
      else
        { 
	  double 
	    irez = inf(z.re()),
	    srez = sup(z.re()),
	    iimz = inf(z.im()),
	    simz = sup(z.im());
	  Interval 
	    hxl( irez ), hxu( srez ), hyl( iimz ), hyu( simz );
	  double 
	    resxl,  resxu, resyl, resyu;
	  CInterval res;
	  //
	  //  extremal values lie in corner points or on intersections of the 
	  //  boundary of z with any ray  arg( w ) = k*Pi/(2*(n-1)) 
	  //
	  //  first check the corners of z
	  //
	  res = power_point( CInterval( hxl, hyl ), n );
	  resxl = inf( res.re() );
	  resxu = sup( res.re() );
	  resyl = inf( res.im() );
	  resyu = sup( res.im() );
	  res = power_point( CInterval( hxu, hyl ), n );
	  update_res( res, resxl, resxu, resyl, resyu );
	  res = power_point( CInterval( hxl, hyu ), n );
	  update_res( res, resxl, resxu, resyl, resyu );
	  res = power_point( CInterval( hxu, hyu ), n );
	  update_res( res, resxl, resxu, resyl, resyu );
	  //
	  //  consider the origin, if it is a boundary point 
	  //  (negative n have been taken care of above)
	  //
	  if( irez * srez <= 0.0 and iimz * simz <= 0.0 and irez * srez * iimz * simz == 0.0 )
	    update_res( CInterval( ZERO_INTERVAL(), ZERO_INTERVAL()), resxl, resxu, resyl, resyu );
	  //
	  //  now check for ray intersections 
	  //  for each quadrant and each boundary, we must consider
	  //  the function values at at most 4 consecutive intersections
	  //
	  //  Ia: lower boundary
	  //  
	  if( iimz != 0.0 )
	    {
	      Interval arg_h = arg( CInterval( z.re(), hyl ) );

	      //  check if there are ray intersections
	      horizontal_check( hyl, iimz, arg_h, irez, srez,
				resxl, resxu, resyl, resyu, n );
	    }
	  //  Ib: upper boundary
	  if( simz != 0.0 )
	    {
	      Interval arg_h = arg( CInterval( z.re(), hyu ) );

	      //  check if there are ray intersections
	      horizontal_check( hyu, simz, arg_h, irez, srez,
				resxl, resxu, resyl, resyu, n );
	    }
	  //  IIa: left boundary
	  if( irez != 0.0 )
	    {
	      Interval arg_h = arg( CInterval( hxl, z.im() ) );

	      //  check if there are ray intersections
	      vertical_check( hxl, irez, arg_h, iimz, simz, 
			      resxl, resxu, resyl, resyu, n );
	    }
	  //  IIb: right boundary
	  if( srez != 0.0 )
	    {
	      Interval arg_h = arg( CInterval(  hxu, z.im() ) );

	      //  check if there are ray intersections
	      vertical_check( hxu, srez, arg_h, iimz, simz, 
			      resxl, resxu, resyl, resyu, n );
	    }
	  return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );
	}
    }
}
//
//-- end power ----------------------------------------------------------------

//-- Pow ------------------------------------------------------------ 040627 --
//
//  Analytic power function for real interval exponent, based on Ln.
//
CInterval CoStLy::Pow( const CInterval& z, const Interval& p ) 
{
  if( inf(z.re()) <= 0.0 and inf(z.im()) <= 0.0 and sup(z.im()) >= 0.0  )
    //  z contains 0 or negative real values
    err_msg( "Pow", z );
  else
    return exp( p * Ln( z ) );
}
//
//-- end Pow ------------------------------------------------------------------

//-- Pow ------------------------------------------------------------ 040627 --
//
//  Analytic power function for complex interval exponent, based on Ln.
//
CInterval CoStLy::Pow( const CInterval& z, const CInterval& p ) 
{
  if( inf(z.re()) <= 0.0 and inf(z.im()) <= 0.0 and sup(z.im()) >= 0.0  )
    //  z contains 0 or negative real values
    err_msg( "Pow", z );
  else
    return exp( p * Ln( z ) );
}
//
//-- end Pow ------------------------------------------------------------------

//-- pow_all -------------------------------------------------------- 040627 --
//
//  Non-analytic power function for real interval exponent.
//
//  If 0 \not\in z, then compute four rectangular intervals that comprehend
//  an annulus which contains all values  zeta^phi, zeta in z, phi in p.
//
//  If 0 in z and negative reals in p, then return the whole complex plane.
//
std::list<CInterval> CoStLy::pow_all( const CInterval& z, const Interval& p )
{
  Interval abs_z( abs( z ) );
  Interval arg_z( arg( z ) );

  if( 0.0 < inf(abs_z) )
    {
      Interval abs_z_p = exp( p * ln( abs_z ) );

      //  Inner and outer radii of the annulus are inf/sup(abs_z_n)
      //  Inscribed square has side length sqrt( 2 ) * rad_1
      Interval rad_1 = INV_SQRT_2() * inf( abs_z_p );
      Interval rad_2 = sup( abs_z_p );

      std::list<CInterval> res;
      
      //  4 intervals covering the annulus, counter-clockwise
      res.push_back( CInterval( Interval(  inf( rad_1 ),  sup( rad_2 ) ),
				Interval( -sup( rad_1 ),  sup( rad_2 ) ) ) );
      res.push_back( CInterval( Interval( -sup( rad_2 ),  sup( rad_1 ) ),
				Interval(  inf( rad_1 ),  sup( rad_2 ) ) ) );
      res.push_back( CInterval( Interval( -sup( rad_2 ), -inf( rad_1 ) ),
				Interval( -sup( rad_2 ),  sup( rad_1 ) ) ) );
      res.push_back( CInterval( Interval( -sup( rad_1 ),  sup( rad_2 ) ),
				Interval( -sup( rad_2 ), -inf( rad_1 ) ) ) );
  
      return res;
    }
  else 
    //  0 in z
    {
      if( inf( p ) < 0.0 )
	//  zeta^phi, zeta in z, phi in p is an unbounded set
	//  return entire complex plane
	{
	  std::list<CInterval> res;
	  
	  res.push_back( CInterval( Interval( -numeric_limits<double>::infinity(), numeric_limits<double>::infinity() ),
				    Interval( -numeric_limits<double>::infinity(), numeric_limits<double>::infinity() ) ) ); 
	  return res;
	}
      else if( inf( p ) == 0.0 and sup( p ) == 0.0 )
	//  z^0 := [0,1]
	{
	  std::list<CInterval> res;
	  
	  res.push_back( CInterval( Interval( 0.0, 1.0 ),
				    Interval( ZERO_INTERVAL() ) ) ); 

	  return res;
	}
      else
	//  All zeta^phi, zeta in z, phi in p  lie in a disk
	{
	  Interval abs_z_p = exp( p * ln( Interval( sup( abs_z ) ) ) );
	  double rad_p = sup( abs_z_p );
	  
	  std::list<CInterval> res;
	  
	  res.push_back( CInterval( Interval( 0.0, rad_p ) * cos( p * arg_z ),
				    Interval( 0.0, rad_p ) * sin( p * arg_z ) ) ); 
	  
	  return res;
	}
    }
}
//
//-- end pow_all --------------------------------------------------------------

//-- pow_all -------------------------------------------------------- 040628 --
//
//  Non-analytic power function for complex interval exponent.
//
//  If p contains nonreal numbers, then the set  zeta^phi, zeta in z, phi in p
//  is unbounded and the whole complex plane is returned.
//
std::list<CInterval> CoStLy::pow_all( const CInterval& z, const CInterval& p )
{
  if( p.im() == ZERO_INTERVAL() ) 
    return pow_all( z, p.re() );
  else 
    //  nonreal exponent: return entire complex plane.
    {
      std::list<CInterval> res;
      
      res.push_back( CInterval( Interval( -numeric_limits<double>::infinity(), numeric_limits<double>::infinity() ),
				Interval( -numeric_limits<double>::infinity(), numeric_limits<double>::infinity() ) ) ); 
      
      return res;
    }
}
//
//-- pow_all ------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Section 5: Asin, Acos, Asinh, Acosh
//
//  The implementation of Acos, Asinh, and Acosh is based on Asin:
//
//  Acos( z )  = pi / 2 - Asin( z )
//  Asinh( z ) = i * Asin( - i * z )
//  Acosh( z ) = i * Acos( z ) = +/- i * ( pi / 2 - Asin( z ) ) 
//
//  Only principal values are computed.
//
//-----------------------------------------------------------------------------

//-- Asin ----------------------------------------------------------- 040905 --
//
//  Analytic inverse sine function
//
Interval f_aux_asin( const Interval& x, const Interval& y ) //------- 040905 --
//
//  auxiliary function for evaluating Asin
//  f_aux_asin(z) = ( |z+1| + |z-1| ) / 2
//
{
  Interval sqr_y = sqr( y );
  Interval res = ( sqrt( sqr( x + 1.0 ) + sqr_y ) + sqrt( sqr( x - 1.0 ) + sqr_y ) ) / 2.0; 
  //  correct overestimation of lower boundary
  //  (x is a point interval)
  double hlb = max( 1.0, std::abs( sup( x ) ) );
  if( inf( res ) < hlb )
    //  invalid overestimation!
    res = Interval( hlb, sup (res ) );
  return res;
}

CInterval CoStLy::Asin( const CInterval& z ) //---------------------- 040730 --
{
  Interval
    rez = z.re(),
    imz = z.im();

  double 
    irez = inf(rez),
    srez = sup(rez),
    iimz = inf(imz),
    simz = sup(imz);

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double 
    resxl, resxu, resyl, resyu;
  //
  //  1st: check for singularities
  //
  if( ( iimz <= 0.0 and simz >= 0.0 ) and ( irez < -1.0 or srez > 1.0 ) ) 
    err_msg( "Asin", z );
  //
  //  2nd: real part
  //
  if( iimz < 0.0 and simz > 0.0 )
    //  z intersects [-1,1]
    {
      if( irez <= 0.0 )
	resxl = inf( asin( hxl ) );
      else
	resxl = inf( asin( hxl / f_aux_asin( hxl, Interval( max( - iimz, simz ) ) ) ) );
      if( srez < 0.0 )
	resxu = sup( asin( hxu / f_aux_asin( hxl, Interval( max( - iimz, simz ) ) ) ) ); 
      else
	resxu = sup( asin( hxu ) );
     }
  else
    {
      if( ( iimz >= 0.0 and irez >= 0.0  ) or ( simz <= 0.0 and irez <= 0.0 ) )
	//  left boundary in quadrants I or III
	//  min( Re( z ) ) in upper left corner
	resxl = inf( asin( hxl / f_aux_asin( hxl, hyu ) ) );
      else
	//  left boundary in quadrants II or IV
	//  min( Re( z ) ) in lower left corner
	resxl = inf( asin( hxl / f_aux_asin( hxl, hyl ) ) );
      if( ( iimz >= 0.0 and srez >= 0.0  ) or ( simz <= 0.0 and srez <= 0.0 ) )
	//  right boundary in quadrants I or III
	//  max( Re( z ) ) in lower right corner
	resxu = sup( asin( hxu / f_aux_asin( hxu, hyl ) ) );
      else
	//  right boundary in quadrants II or IV
	//  max( Re( z ) ) in upper right corner
	resxu = sup( asin( hxu / f_aux_asin( hxu, hyu ) ) );
    }
  //
  //  3rd: imaginary part
  //
  if( simz <= 0.0 )
    //  z in lower half plane
    //  min( Im( z ) ) in point with max |z|
    //  max( Im( z ) ) in point with min |z|
    {
      if( irez + srez < 0.0 )
	//  most of z in quadrant III
	{
	  resyl = - sup( acosh( f_aux_asin( hxl, hyl ) ) );
	  if( srez < 0.0 )
	    resyu = - inf( acosh( f_aux_asin( hxu, hyu ) ) );
	  else
	    resyu = - inf( acosh( f_aux_asin( ZERO_INTERVAL(), hyu ) ) );
	}
      else
	//  most of z in quadrant IV
	{
	  resyl = - sup( acosh( f_aux_asin( hxu, hyl ) ) );
	  if( irez > 0.0 )
	    resyu = - inf( acosh( f_aux_asin( hxl, hyu ) ) );
	  else
	    resyu = - inf( acosh( f_aux_asin( ZERO_INTERVAL(), hyu ) ) );
	}
    }
  else if( iimz >= 0.0 )
    //  z in upper half plane
    //  min( Im( z ) ) in point with min |z|
    //  max( Im( z ) ) in point with max |z|
    {
      if( irez + srez < 0.0 )
	//  most of z in quadrant II
	{
	  resyu = sup( acosh( f_aux_asin( hxl, hyu ) ) );
	  if( srez < 0.0 )
	    resyl = inf( acosh( f_aux_asin( hxu, hyl ) ) );
	  else
	    resyl = inf( acosh( f_aux_asin( ZERO_INTERVAL(), hyl ) ) );
	}
      else
	//  most of z in quadrant I
	{
	  resyu = sup( acosh( f_aux_asin( hxu, hyu ) ) );
	  if( irez > 0.0 )
	    resyl = inf( acosh( f_aux_asin( hxl, hyl ) ) );
	  else
	    resyl = inf( acosh( f_aux_asin( ZERO_INTERVAL(), hyl ) ) );
	}
    }
  else
    //  z intersects imaginary axes
    //  min( Im( z ) ) in point in lower half plane with max |z|
    //  max( Im( z ) ) in point in upper half plane with max |z|
    {
      if( irez + srez < 0.0 )
	//  most of z in quadrants II and IV
	{
	  resyl = - sup( acosh( f_aux_asin( hxl, hyl ) ) );
	  resyu = sup( acosh( f_aux_asin( hxl, hyu ) ) );
	}
      else
	{
	  resyl = - sup( acosh( f_aux_asin( hxu, hyl ) ) );
	  resyu = sup( acosh( f_aux_asin( hxu, hyu ) ) );
	}
    }

  return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );

}
//
//-- end Asin -----------------------------------------------------------------

//-- Acos ----------------------------------------------------------- 040730 --
//
CInterval CoStLy::Acos( const CInterval& z ) 
{
  return HALFPI() - Asin( z );
}
//
//-- end Acos -----------------------------------------------------------------

//-- Asinh ---------------------------------------------------------- 040730 --
//
CInterval CoStLy::Asinh( const CInterval& z )
//
//  Asinh( Z ) = i * Asin( -i * z )
//
{
  CInterval res = Asin( CInterval( z.im(), -z.re() ) );
  return CInterval( - res.im(), res.re() );
}
//
//-- end Asinh ----------------------------------------------------------------

//-- Acosh ---------------------------------------------------------- 040908 --
//
CInterval CoStLy::Acosh( const CInterval& z ) 
//
//  Acosh( z ) = i * Acos( z ) = +/- i * ( pi / 2 - Asin( z ) ) 
//
{
  Interval
    rez = z.re(),
    imz = z.im();

  double 
    irez = inf(rez),
    srez = sup(rez),
    iimz = inf(imz),
    simz = sup(imz);

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double 
    resxl, resxu, resyl, resyu;

  //
  //  1st: check for singularities
  //
  if( ( iimz <= 0.0 and simz >= 0.0 ) and ( irez < 1.0 ) ) 
    err_msg( "Acosh", z );
  //
  //  2nd: z in upper half plane
  //  Acosh( z ) =  + i * ( pi / 2 - Asin( z ) ) 
  //
  if( iimz > 0.0 )
    {
      CInterval res = HALFPI() - Asin( z );
      return CInterval( - res.im(), res.re() );
    }
  //
  //  3rd: z in lower half plane
  //  Acosh( z ) =  - i * ( pi / 2 - Asin( z ) ) 
  //
  if( simz < 0.0 )
    {
      CInterval res = HALFPI() - Asin( z );
      return CInterval( res.im(), - res.re() );
    }
  //
  //  z intersects [1,infinity)
  //
  //  real part
  //  minmum on the left on real axes, maximum in lower or upper right corner
  //
  resxl = inf( acosh( hxl ) );
  Interval ytilde( max( -iimz, simz ) );
  resxu = sup( acosh( f_aux_asin( hxu, ytilde ) ) );
  //
  //  imaginary part
  //  minmum in lower left corner, maximum in upper left corner
  //
  resyl = - sup( acos( hxl / f_aux_asin( hxl, hyl ) ) );
  resyu = sup( acos( hxl / f_aux_asin( hxl, hyu ) ) );
  
  return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );

}
//
//-- end Acosh ----------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//  Section 6: Atan, Acot, Atanh, Acoth
//
//  The implementation of Acot, Atanh, and Acoth is based on Atan:
//
//  Acot( z )  =  Atan( 1 / z )
//  Atanh( z ) = - i * Atan( i * z )
//  Acoth( z ) = i * Acot( i * z )
//
//  Only principal values are computed.
//
//-----------------------------------------------------------------------------

//-- Atan ----------------------------------------------------------- 040912 --
//
//  Analytic inverse tangent function
//
void re_vert( const double& x, const Interval& hx, 
	      const double& rew_inf, const double& rew_sup,
 	      double& resxl, double& resxu ) //---------------------- 040729 --
//
//  Subroutine of analytic inverse tangent function.
//  Evaluate real part on a vertical boundary.
//
{
  if( x == 0.0 )
    //  singularities have been handled before, hence Re( w ) > 0
    {
      resxl = 0.0;
      resxu = 0.0;
    }
  else
    { 
      if( x > 0.0 )
	//  w in quadrants I and/or II
	//  atan is the inverse function of tan(t), t in (-pi/2,pi/2).
	{
	  resxl = ( rew_sup > 0.0 ? inf( atan( 2 * hx / rew_sup ) ) 
                        : ( rew_sup < 0.0 ? inf( atan( 2*hx/rew_sup ) + PI() )
		                          : inf( HALFPI() ) ) ) / 2.0;
	  resxu = ( rew_inf > 0.0 ? sup( atan( 2 * hx / rew_inf ) ) 
		        : ( rew_inf < 0.0 ? sup( atan( 2*hx/rew_inf ) + PI() )
			                  : sup( HALFPI() ) ) ) / 2.0;
	}
      else
	//  w in quadrants III and/or IV
	{
	  resxl = ( rew_inf < 0.0 ? inf( atan( 2 * hx / rew_inf ) - PI() )
                            : ( rew_inf > 0.0 ? inf( atan( 2*hx/rew_inf ) )
			                      : - sup( HALFPI() ) ) ) / 2.0;
	  resxu = ( rew_sup < 0.0 ? sup( atan( 2 * hx / rew_sup ) - PI() )
                            : ( rew_sup > 0.0 ? sup( atan( 2*hx/rew_sup ) )
			                      : - inf( HALFPI() ) ) ) / 2.0;
	}
    }
}

CInterval CoStLy::Atan( const CInterval& z ) //---------------------- 040912 --
{
  Interval
    rez = z.re(),
    imz = z.im();

  double 
    irez = inf(rez),
    srez = sup(rez),
    iimz = inf(imz),
    simz = sup(imz);

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double 
    resxl, resxu, resyl, resyu;
  //
  //  1st: check for singularities
  //
  if( ( irez <= 0.0 and srez >= 0.0 ) and ( iimz <= -1.0 or simz >= 1.0 ) ) 
    err_msg( "Atan", z );
  //
  //  2nd: real part
  //  Re( Atan( z ) ) = Arg( w ) / 2, where w = 1 - x^2 - y^2 + i * 2x ) 
  //
  //  evaluate Atan on vertical boundaries
  //
  Interval
    y_sqr = sqr( imz ),
    rew_l = 1 - sqr( hxl ) - y_sqr,
    rew_u = 1 - sqr( hxu ) - y_sqr;
  //
  //  left boundary
  //
  double rew_inf = inf( rew_l );
  double rew_sup = sup( rew_l );
  re_vert( irez, hxl, rew_inf, rew_sup, resxl, resxu );
  //
  //  right boundary
  //
  rew_inf = inf( rew_u );
  rew_sup = sup( rew_u );
  double res_l, res_u;
  re_vert( srez, hxu, rew_inf, rew_sup, res_l, res_u );

  resxl = min( resxl, res_l );
  resxu = max( resxu, res_u );
  //
  //  look for extremal values on horizontal boundaries
  //  since Atan( x+iy ) = Atan( x-iy ), 
  //  intersections can be considered in the upper half plane
  //
  double abs_y_min = inf( abs( imz ) );

  if( abs_y_min > 1.0 )
    {
      Interval 
	abs_hyl = Interval( abs_y_min ),
	abs_hxl = sqrt( sqr( abs_hyl ) - 1.0 );

      if( sup( abs_hxl ) > irez and inf( abs_hxl ) < srez )
	//  extremal curve intersects lower boundary of x+i|y| in quadrant I
	//  intersection in Q I or Q IV: update minimum
	//	resxl = inf( atan( abs_y_min / abs_hxl ) ) / 2.0;
	resxl = inf( PI() - atan( 1.0 / abs_hxl ) ) / 2.0;
      else if( - inf( abs_hxl ) > irez and - sup( abs_hxl ) < srez )
	//  extremal curve intersects lower boundary of x+i|y| in quadrant II
	//  intersection in Q II or Q III: update maximum
	resxu = sup( atan( 1.0 / abs_hxl ) - PI() ) / 2.0;
    }
  //
  //  3rd: imaginary part
  //  Im( Atan( z ) ) = +/- Ln( 1 +/- 4y/( x^2 + (1 -/+ y)^2 ) ) / 4 
  //
  //  evaluate atan on horizontal boundaries
  //
  Interval
    x_sqr = sqr( rez ), 
    im_atan_l, im_atan_u;

  if( iimz < 0.0 )
    im_atan_l = - ln( 1 - 4 * hyl / ( x_sqr + sqr( 1 + hyl ) ) ) / 4.0;
  else
    im_atan_l = ln( 1 + 4 * hyl / ( x_sqr + sqr( 1 - hyl ) ) ) / 4.0;
  if( simz < 0.0 )
    im_atan_u = - ln( 1 - 4 * hyu / ( x_sqr + sqr( 1 + hyu ) ) ) / 4.0;
  else
    im_atan_u = ln( 1 + 4 * hyu / ( x_sqr + sqr( 1 - hyu ) ) ) / 4.0;

  resyl = min( inf( im_atan_l ), inf( im_atan_u ) );
  resyu = max( sup( im_atan_l ), sup( im_atan_u ) );
  //
  //  look for extremal values on vertical boundaries,
  //  if vertical boundaries intersect extremal curves
  //
  double abs_x_min = inf( abs( rez ) );
  Interval 
    x_extr = Interval( abs_x_min ),
    y_extr = sqrt( 1.0 + sqr( x_extr ) );

  if( inf( y_extr ) < simz and sup( y_extr ) > iimz )
    //  extremal curve intersects left boundary of |x|+iy in quadrant I
    //  update maximum
    resyu = sup( ln( 1 + 4 * y_extr / ( sqr( x_extr ) + sqr( 1 - y_extr ) ) ) ) / 4.0;
  if( - sup( y_extr ) < simz and - inf( y_extr ) > iimz )
    //  extremal curve intersects left boundary of |x|+iy in quadrant IV
    //  update minimum
    resyl = - sup( ln( 1 + 4 * y_extr / ( sqr( x_extr ) + sqr( 1 - y_extr ) ) ) ) / 4.0;

  return CInterval( Interval( resxl, resxu ), Interval( resyl, resyu ) );

}
//
//-- end Atan -----------------------------------------------------------------

//-- Acot ----------------------------------------------------------- 040912 --
//
//  Analytic inverse cotangent function
//  Acot( z ) = Atan( 1/z )
//  The code of Acot( z ) is almost identical to the code of Atan( z ) 
//
CInterval CoStLy::Acot( const CInterval& z )
{
  Interval
    rez = z.re(),
    imz = z.im();

  double 
    irez = inf(rez),
    srez = sup(rez),
    iimz = inf(imz),
    simz = sup(imz);

  Interval 
    hxl(irez), hxu(srez), hyl(iimz), hyu(simz);

  double 
    resxl, resxu, resyl, resyu;
  //
  //  1st: check for singularities
  //
  if( ( irez <= 0.0 and srez >= 0.0 ) and ( iimz <= 1.0 and simz >= -1.0 ) ) 
    err_msg( "Acot", z );
  //
  //  2nd: real part
  //  Re( Atan(  z  ) )   = Arg( w ) / 2, where w = 1 - x^2 - y^2 + i * 2x ) 
  //  Re( Atan( 1 / z ) ) = Arg( w ) / 2, where w = x^2 + y^2 - 1 + i * 2x ) 
  //
  //  evaluate Acot on vertical boundaries
  //
  Interval
    y_sqr = sqr( imz ),
    rew_l = sqr( hxl ) + y_sqr - 1,
    rew_u = sqr( hxu ) + y_sqr - 1;
  //
  //  left boundary
  //
  double rew_inf = inf( rew_l );
  double rew_sup = sup( rew_l );
  re_vert( irez, hxl, rew_inf, rew_sup, resxl, resxu );
  //
  //  right boundary
  //
  rew_inf = inf( rew_u );
  rew_sup = sup( rew_u );
  double res_l, res_u;
  re_vert( srez, hxu, rew_inf, rew_sup, res_l, res_u );

  resxl = min( resxl, res_l );
  resxu = max( resxu, res_u );
  //
  //  look for extremal values on horizontal boundaries
  //  since Acot( x+iy ) = Acot( x-iy ), 
  //  intersections can be considered in the upper half plane
  //
  double abs_y_min = inf( abs( imz ) );

  if( abs_y_min > 1.0 )
    {
      Interval 
	abs_hyl = Interval( abs_y_min ),
	abs_hxl = sqrt( sqr( abs_hyl ) - 1.0 );

      if( sup( abs_hxl ) > irez and inf( abs_hxl ) < srez )
	//  extremal curve intersects lower boundary of x+i|y| in quadrant I
	//  intersection in Q I or Q IV: update maximum
	resxu = sup( atan( 1.0 / abs_hxl ) ) / 2.0;
      if( - inf( abs_hxl ) > irez and - sup( abs_hxl ) < srez )
	//  extremal curve intersects lower boundary of x+i|y| in quadrant II
	//  intersection in Q II or Q III: update minimum
	resxl = - sup( atan( 1.0 / abs_hxl ) ) / 2.0; 
     }
  //
  //  3rd: imaginary part
  //  Im( Atan( z ) ) = +/- Ln( 1 +/- 4y/( x^2 + (1 -/+ y)^2 ) ) / 4 
  //  Im( Acot ) = - Im ( Atan ): We calculate Im( Atan ) and return "-"
  //
  //  evaluate atan on horizontal boundaries
  //
  Interval
    x_sqr = sqr( rez ), 
    im_atan_l, im_atan_u;

  if( iimz < 0.0 )
    im_atan_l = - ln( 1 - 4 * hyl / ( x_sqr + sqr( 1 + hyl ) ) ) / 4.0;
  else
    im_atan_l = ln( 1 + 4 * hyl / ( x_sqr + sqr( 1 - hyl ) ) ) / 4.0;
  if( simz < 0.0 )
    im_atan_u = - ln( 1 - 4 * hyu / ( x_sqr + sqr( 1 + hyu ) ) ) / 4.0;
  else
    im_atan_u = ln( 1 + 4 * hyu / ( x_sqr + sqr( 1 - hyu ) ) ) / 4.0;

  resyl = min( inf( im_atan_l ), inf( im_atan_u ) );
  resyu = max( sup( im_atan_l ), sup( im_atan_u ) );
  //
  //  look for extremal values on vertical boundaries,
  //  if vertical boundaries intersect extremal curves
  //
  double abs_x_min = inf( abs( rez ) );
  Interval 
    x_extr = Interval( abs_x_min ),
    y_extr = sqrt( 1.0 + sqr( x_extr ) );

  if( inf( y_extr ) < simz and sup( y_extr ) > iimz )
    //  extremal curve intersects left boundary of |x|+iy in quadrant I
    //  update maximum
    resyu = sup( ln( 1 + 4 * y_extr / ( sqr( x_extr ) + sqr( 1 - y_extr ) ) ) ) / 4.0;
  if( - sup( y_extr ) < simz and - inf( y_extr ) > iimz )
    //  extremal curve intersects left boundary of |x|+iy in quadrant IV
    //  update minimum
    resyl = - sup( ln( 1 + 4 * y_extr / ( sqr( x_extr ) + sqr( 1 - y_extr ) ) ) ) / 4.0;

  return CInterval( Interval( resxl, resxu ), Interval( - resyu, - resyl ) );

}
//
//-- end Acot -----------------------------------------------------------------

//-- Atanh ---------------------------------------------------------- 040912 --
//
CInterval CoStLy::Atanh( const CInterval& z )
//
//  Atanh( z ) = - i * Atan( i * z )
//
{
  CInterval res = Atan( CInterval( - z.im(), z.re() ) );
  return CInterval( res.im(), - res.re() );
}
//
//-- end Atanh ----------------------------------------------------------------

//-- Acoth ---------------------------------------------------------- 040912 --
//
CInterval CoStLy::Acoth( const CInterval& z ) 
//
//  Acoth( z ) = i * Acot( i * z )
//
{
  CInterval res = Acot(  CInterval( - z.im(), z.re() ) );
  return CInterval( - res.im(), res.re() );
}
//
//-- end Acoth ----------------------------------------------------------------

/*

  End of File: cimath.cpp

*/
