/*

 File: taylor.tpl, 2004/06/14

 ACETAF, Version 3.0.0

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <sstream>

#include "error.h"
#include "zero.h"
#include "functions.tpl" //Function templates.

/*
  Ctors
*/

template<class T> Taylor<T>::Taylor(int v, unsigned int order, unsigned int cap) 
  : cont_( order + 1, T(0.0) )
{
  if( cap > order + 1 ) cont_.reserve(cap);
  
  cont_[0] = T(v);
  if( order > 0 ) cont_[1] = T(1.0);
}
template<class T> Taylor<T>::Taylor(const double& v, unsigned int order, unsigned int cap) 
  : cont_( order + 1, T(0.0) )
{
  if( cap > order + 1 ) cont_.reserve(cap);
  
  cont_[0] = T(v);
  if( order > 0 ) cont_[1] = T(1.0);
}
template<class T> Taylor<T>::Taylor(const T& v, unsigned int order, unsigned int cap) 
  : cont_( order + 1, T(0.0) )
{
  if( cap > order + 1 ) cont_.reserve(cap);
  
  cont_[0] = v;
  if( order > 0 ) cont_[1] = T(1.0);
}

/*
  Arithmetic operations.
*/

template<class T> Taylor<T> Taylor<T>::operator - () const
{
  Taylor<T> result;
  result.cont_.resize( cont_.size() );
  for(int i = 0; i < cont_.size(); i++) result.cont_[i] = -cont_[i];

  return result;
}

template<class T> Taylor<T>& Taylor<T>::operator += (const Taylor<T>& x)
{
  int thiso = order(), xo = x.order();  

  if( thiso != xo && xo != 0 && thiso != 0) 
    throw wrong_dimensions();//Addition error: orders don't go together.
  
  if( thiso != xo )
    {
      if( thiso == 0 )
	{
	  T tmp( cont_[0] );
	  cont_ = x.cont_;
	  cont_[0] += tmp;

	  return *this;
	}
      else 
	{
	  cont_[0] += x.cont_[0];
	  return *this;
	}
    }
  else
    {
      for(int i=0; i <= xo; i++) cont_[i] += x.cont_[i];
      return *this;
    }
}

template<class T> Taylor<T>& Taylor<T>::operator -= (const Taylor<T>& x)
{
  int thiso = order(), xo = x.order();

  if( thiso != xo && xo != 0 && thiso != 0 ) 
    throw wrong_dimensions();//Subtraction error: orders don't go together.

  if( xo == 0 )
    {
      cont_[0] -= x.cont_[0];
      return *this;
    }
  else
    {
      if( thiso == 0 ) cont_.resize( xo+1, T(0.0) );
      for(int i=0; i <= xo; i++) cont_[i] -= x.cont_[i];
      return *this;
    }
}

template<class T> struct dotprec {}; //dummy-type

template<class T> struct is_zero<Taylor<T> > : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x) const { return is_zero<T>()( x[0] ); }
};

template<class T> Taylor<T>& Taylor<T>::operator *= (const Taylor<T>& x)
{
  int thiso = order(), xo = x.order();

  if( thiso != xo && xo != 0 && thiso != 0) 
    throw wrong_dimensions();//Multiplication error: Orders don't go together.

  if( thiso != xo )
    {
      T tmp;
      int ord;

      if( thiso == 0 )
	{
	  tmp = cont_[0];
	  cont_ = x.cont_;
	  ord = xo;
	}
      else
	{
	  tmp = x.cont_[0];
	  ord = thiso;
	}

      for(int j=0; j<=ord; j++) cont_[j] *= tmp;
	
      return *this;
    }
  else
    {
      std::vector<T> tmp( cont_ );

      for(int j=0; j<=xo; j++) 
	{
	  cont_[j] = T(0.0);
	  for(int i=0; i<=j; i++) cont_[j] += tmp[i] * x.cont_[j-i];
	}
      return *this;
    }
}

template<class T> Taylor<T>& Taylor<T>::operator /= (const Taylor<T>& x)
{
  int thiso = order(), xo = x.order();

  if( thiso != xo && xo != 0 && thiso != 0) 
    throw wrong_dimensions();//Division error: Orders don't go together.

  if( is_zero<Taylor<T> >()(x) ) throw division_by_zero();//Division by 0.

  T x0( 1/x.cont_[0] );
  if( xo == 0 )
    {
      for(int j=0; j<=thiso; j++) cont_[j] *= x0;
      return *this;
    }
  else
    {
      if( thiso == 0 ) cont_.resize( xo+1, T(0.0) );

      std::vector<T> tmp( cont_ );

      cont_[0] *= x0;
      for(int j=1; j<=xo; j++) 
	{
          cont_[j] = T(0.0);
	  for(int i=1; i<=j; i++)
	    {
	      cont_[j] += x.cont_[i] * cont_[j-i];
	    }
	  cont_[j] = ( tmp[j] - cont_[j] ) * x0;
	}

      return *this;
    }
}
  
template<class T> Taylor<T>& Taylor<T>::operator += (const T& d)
{
  cont_[0] += d;
  return *this;
}
template<class T> Taylor<T>& Taylor<T>::operator -= (const T& d)
{
  cont_[0] -= d;
  return *this;
}
template<class T> Taylor<T>& Taylor<T>::operator *= (const T& d)
{
  for(int j=0; j <= order(); j++) cont_[j] *= d;
  return *this;
}
template<class T> Taylor<T>& Taylor<T>::operator /= (const T& d)
{
  if( is_zero<T>()(d) ) throw division_by_zero();

  for(int j=0; j <= order(); j++) cont_[j] /= d;
  return *this;
}

/*
  Definition of friend functions.
*/

enum { _ln,

       _tan,
       _cot,
       
       _asin,
       _acos,
       _atan,
       _acot,
       
       _tanh,
       _coth,
       
       _asinh,
       _acosh,
       _atanh,
       _acoth };

//Auxiliary function

template<class T> void f_g_u(Taylor<T>& f,const Taylor<T>& g,const Taylor<T>& u,int nb_function)
{
  switch( nb_function ) //0-th element
    {
    case _ln:    f.cont_[0] = LN<T>(    u.cont_[0] ); break;
    case _tan:   f.cont_[0] = TAN<T>(   u.cont_[0] ); break;
    case _cot:   f.cont_[0] = COT<T>(   u.cont_[0] ); break;
    case _asin:  f.cont_[0] = ASIN<T>(  u.cont_[0] ); break;
    case _acos:  f.cont_[0] = ACOS<T>(  u.cont_[0] ); break;
    case _atan:  f.cont_[0] = ATAN<T>(  u.cont_[0] ); break;
    case _acot:  f.cont_[0] = ACOT<T>(  u.cont_[0] ); break;
    case _tanh:  f.cont_[0] = TANH<T>(  u.cont_[0] ); break;
    case _coth:  f.cont_[0] = COTH<T>(  u.cont_[0] ); break; 
    case _asinh: f.cont_[0] = ASINH<T>( u.cont_[0] ); break;
    case _acosh: f.cont_[0] = ACOSH<T>( u.cont_[0] ); break;
    case _atanh: f.cont_[0] = ATANH<T>( u.cont_[0] ); break;
    case _acoth: f.cont_[0] = ACOTH<T>( u.cont_[0] ); break;  
    }

  //Remaining elements:
  int size( f.order() );
  T g0( 1/g.cont_[0] );
  for(int j=1; j<=size; j++) 
    {
      T akku(0.0);
      for(int i=1; i<=j-1; i++)
	{
	  akku += i * f.cont_[i] * g.cont_[j-i];
	}
      f.cont_[j] = ( u.cont_[j] - akku/j )*g0;
    }
}

template<class T> Taylor<T> sqrt (const Taylor<T>& x)
{
  if( is_zero<Taylor<T> >()(x) ) throw function_not_defined();//Error in sqrt: Zero argument. See next comment.

  int xo = x.order();
  
  Taylor<T> result( SQRT<T>( x.cont_[0] ), xo );

  if( xo == 0 ) return result;

  T factor( T(0.5)/result.cont_[0] );//The zero argument would cause an error at this point.
  for(int j=1; j<=xo; j++) 
    {
      T akku(0.0);
      for(int i=1; i<=j-1; i++)
	{
	  akku += result.cont_[i] * result.cont_[j-i];
	}
      result.cont_[j] = factor*( x.cont_[j] - akku );
    }
  return result;
}

template<class T> Taylor<T> sqrt (const Taylor<T>& x,int n)
{
  if     ( n == 0 ) return Taylor<T>( T(1.0), 0 );
  else if( n == 1 ) return x;
  else if( n == 2 ) return sqrt( x );
  else return pow( x , T(1.0)/n );
}

template<class T> Taylor<T> sqr (const Taylor<T>& x)
{
  int xo = x.order();
  
  Taylor<T> result( SQR<T>( x.cont_[0] ), xo );
  
  if( xo == 0 ) return result;
  
  //'result.cont_[1]' has been set to 1.0 (two lines above).
  result.cont_[1] = T(0.0);

  for(int j=1; j<=xo; j++) 
    {
      for(int i=0; i<=j; i++) result.cont_[j] += x.cont_[i] * x.cont_[j-i];
    }

  return result;
}


template<class T> Taylor<T> pow (const Taylor<T>& x,const T& alpha)
{
  int xo = x.order();

  Taylor<T> result( POW<T>( x.cont_[0], alpha ), xo );

  if( xo == 0 ) return result;

  if( is_zero<Taylor<T> >()(x) ) throw function_not_defined();//Error in pow: Division by zero in next line. 
      
  T x0( 1/x.cont_[0] );
  for(int j=1; j<=xo; j++) 
    {
      T akku(0.0);
      for(int i=0; i<=j-1; i++)
	{
	  akku += ( alpha*( j-i )- i ) * x.cont_[j-i] * result.cont_[i];
	}
      result.cont_[j] =  akku/j * x0;
    }
  return result; 
}

template<class T> Taylor<T> pow (const Taylor<T>& x,const Taylor<T>& y)
{
  return exp( y * ln( x ) );
}

template<class T> Taylor<T> power (const Taylor<T>& x,int n)
{
  if     ( n <  0 ) return T(1.0) / power( x, -n );
  else if( n == 0 ) return Taylor<T>( T(1.0), 0 );
  else if( n == 1 ) return x;
  else if( n == 2 ) return sqr( x );
  else
    {
      Taylor<T> result( x ), u( T(1.0) );
      while( n > 0 )
        {
	  if( (n & 1) == 0 )//even exponent
       	    {
	      result = sqr( result );
	      n >>= 1;     //Division by 2, realized with a right shift
	    }
	  else
	    {	
              u *= result;
	      n -= 1;
	    }
        }
      return u;
    }
}

template<class T> Taylor<T> exp (const Taylor<T>& x)
{
  int xo = x.order();

  Taylor<T> result( EXP<T>( x.cont_[0] ), xo );

  if( xo == 0 ) return result;
  
  for(int j=1; j<=xo; j++) 
    {
      T akku(0.0);
      for(int i=0; i<=j-1; i++)
	{
	  akku += (j-i) * result.cont_[i] * x.cont_[j-i];
	}
      result.cont_[j] = akku/j;
    }
  return result; 
}

//For argument check:
template<class T,class S> struct arg_check : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x) { return S()(x); }
};

template<class T> struct LOG_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ASIN_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ACOS_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ASINH_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ACOSH_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ATANH_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};

template<class T> struct ACOTH_DEF : public std::unary_function<Taylor<T>,bool>
{
  bool operator () (const Taylor<T>& x);
};


template<class T> Taylor<T> ln (const Taylor<T>& x)
{
  if( arg_check<T,LOG_DEF<T> >()( x ) ) throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );

  f_g_u( f, x, x, _ln);
  
  return f;
}
  
template<class T> Taylor<T> sin (const Taylor<T>& x)
{
  int xo = x.order();

  if( xo == 0 )
    {
      return Taylor<T>( SIN<T>( x.cont_[0] ), xo );
    }
  else
    {
      Taylor<T> result1( SIN<T>( x.cont_[0] ), xo ); //sin
      Taylor<T> result2( COS<T>( x.cont_[0] ), xo ); //cos
    
      for(int j=1; j<=xo; j++) 
	{
	  T akku1(0.0), akku2(0.0);
	  for(int i=0; i<=j-1; i++)
	    {
	      akku1 += (j-i) * result2.cont_[i] * x.cont_[j-i];
	      akku2 += (j-i) * result1.cont_[i] * x.cont_[j-i];
	    }
	  result1.cont_[j] = akku1/j;
	  result2.cont_[j] = -akku2/j;
	}
      return result1; 
    }
}

template<class T> Taylor<T> cos (const Taylor<T>& x)
{
  int xo = x.order();

  if( xo == 0 )
    {
      return Taylor<T>( COS<T>( x.cont_[0] ), xo ); 
    }
  else
    {
      Taylor<T> result1( SIN<T>( x.cont_[0] ), xo ); //sin
      Taylor<T> result2( COS<T>( x.cont_[0] ), xo ); //cos
   
      for(int j=1; j<=xo; j++) 
	{
	  T akku1(0.0), akku2(0.0);
	  for(int i=0; i<=j-1; i++)
	    {
	      akku1 += (j-i) * result2.cont_[i] * x.cont_[j-i];
	      akku2 += (j-i) * result1.cont_[i] * x.cont_[j-i];
	    }
	  result1.cont_[j] = akku1/j;
	  result2.cont_[j] = -akku2/j;
	}
      return result2; 
    }
}

template<class T> Taylor<T> tan (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( sqr(cos(x)) );
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _tan);
  
  return f;
}

template<class T> Taylor<T> cot (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( -sqr(sin(x)) ); 
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _cot);
  
  return f;
}


template<class T> Taylor<T> sinh (const Taylor<T>& x)
{
  int xo = x.order();

  if( xo == 0 )
    {
      return Taylor<T>( SINH<T>( x.cont_[0] ), xo );
    }
  else
    {
      Taylor<T> result1( SINH<T>( x.cont_[0] ), xo ); //sinh
      Taylor<T> result2( COSH<T>( x.cont_[0] ), xo ); //cosh
   
      for(int j=1; j<=xo; j++) 
	{
	  T akku1(0.0), akku2(0.0);
	  for(int i=0; i<=j-1; i++)
	    {
	      akku1 += (j-i) * result2.cont_[i] * x.cont_[j-i];
	      akku2 += (j-i) * result1.cont_[i] * x.cont_[j-i];
	    }
	  result1.cont_[j] = akku1/j;
	  result2.cont_[j] = akku2/j;
	}
      return result1; 
    }
}

template<class T> Taylor<T> cosh (const Taylor<T>& x)
{
  int xo = x.order();

  if( xo == 0 )
    {
      return Taylor<T>( COSH<T>( x.cont_[0] ), xo );
    }
  else
    {
      Taylor<T> result1( SINH<T>( x.cont_[0] ), xo ); //sinh
      Taylor<T> result2( COSH<T>( x.cont_[0] ), xo ); //cosh
  
      for(int j=1; j<=xo; j++) 
	{
	  T akku1(0.0), akku2(0.0);
	  for(int i=0; i<=j-1; i++)
	    {
	      akku1 += (j-i) * result2.cont_[i] * x.cont_[j-i];
	      akku2 += (j-i) * result1.cont_[i] * x.cont_[j-i];
	    }
	  result1.cont_[j] = akku1/j;
	  result2.cont_[j] = akku2/j;
	}
      return result2; 
    }
}

template<class T> Taylor<T> tanh (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( sqr(cosh(x)) );
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _tanh);
  
  return f;
}

template<class T> Taylor<T> coth (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( -sqr(sinh(x)) );
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _coth);
  
  return f;
}

  
template<class T> Taylor<T> asin (const Taylor<T>& x)
{
  if( arg_check<T,ASIN_DEF<T> >()( x ) ) 
    throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( sqrt( T(1.0) - sqr(x) ));
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _asin);
  
  return f;
}

template<class T> Taylor<T> acos (const Taylor<T>& x)
{
  if( arg_check<T,ACOS_DEF<T> >()( x ) )
    throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( -sqrt( T(1.0) - sqr(x) ) );
    
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _acos);
  
  return f;
}

template<class T> Taylor<T> atan (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( T(1.0) + sqr(x) );

  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
 
  f_g_u( f, g, x, _atan);
  
  return f;
}

template<class T> Taylor<T> acot (const Taylor<T>& x)
{
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( -( T(1.0) + sqr(x)) );
  
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.

  f_g_u( f, g, x, _acot);
  
  return f;
}


template<class T> Taylor<T> asinh (const Taylor<T>& x)
{
  if( arg_check<T,ASINH_DEF<T> >()( x ) )
    throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( sqrt( T(1.0) + sqr(x) ) );
    
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
 
  f_g_u( f, g, x, _asinh);
  
  return f;
}

template<class T> Taylor<T> acosh (const Taylor<T>& x)
{
  if( arg_check<T,ACOSH_DEF<T> >()( x ) )
    throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( sqrt( sqr(x) - T(1.0) ) );
    
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _acosh);
  
  return f;
}

template<class T> Taylor<T> atanh (const Taylor<T>& x)
{
  if( arg_check<T,ATANH_DEF<T> >()( x ) )
    throw function_not_defined();
  
  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( T(1.0) - sqr(x) );
    
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _atanh);
  
  return f;
}

template<class T> Taylor<T> acoth (const Taylor<T>& x)
{
  if( arg_check<T,ACOTH_DEF<T> >()( x ) )
    throw function_not_defined();

  Taylor<T> f( T(0.0), x.order() );
  Taylor<T> g( T(1.0) - sqr(x) );
    
  if( is_zero<Taylor<T> >()( g ) ) throw division_by_zero();//Zero argument causes div. by 0 in f_g_u.
  
  f_g_u( f, g, x, _acoth);
  
  return f;
}

template<class T> std::ostream& operator << (std::ostream& os, const Taylor<T>& x)
{
  for(unsigned int i=0; i <= x.order(); i++) 
    {
      os << i << ": " << x.cont_[i] << "\n";
    }
  return os;
}
template<class T> std::istream& operator >> (std::istream& s, Taylor<T>& x)
{
  if( x.cont_.empty() ) x.cont_.resize(1);
  s >> x.cont_[0];
}
template<class T> std::string& operator << (std::string& s,const Taylor<T>& x)
{
  std::ostringstream out;
  out << x;

  return (s = out.str()); 
}
template<class T> std::string& operator >> (std::string& s,     Taylor<T>& x)
{
  std::istringstream in(s);
  in  >> x;

  return (s = in. str()); 
}

/*

  End of File: taylor.tpl

*/
